from iwisdm.core import Env
from iwisdm.envs.registration import StimData, EnvSpec
from iwisdm.envs.shapenet.shapenet_env import ShapeNetEnv
from iwisdm.utils.read_write import find_data_folder


def make(
        env_id: str = 'ShapeNet',
        stim_data: StimData = None,
        env_spec: EnvSpec = None,
        dataset_fp: str = None,
) -> Env:
    """
    create an environment instance with
    1. specified stimuli dataset, the dataset directory contains the actual stimuli,
        and a meta info file that contains information about the stimuli
        see iwisdm/envs/registration.py for more details on the stimuli dataset
    2. environment specifications that control how tasks are randomly generated
        see iwisdm/envs/registration.py for more details on the environment specification
    @param env_id: the name of the environment
    @param stim_data: Data class that contains the stimuli dataset directory
    @param env_spec: Data class that contains the environment specification
    @param dataset_fp: the file path to the stimuli dataset
    @return:
    """
    assert env_id in env_dict, f"environment {env_id} not found in env_dict"
    env = env_dict[env_id]
    if stim_data is None:
        if dataset_fp is None:
            dataset_fp = find_data_folder()
        stim_data = env.init_stim_data(dataset_fp)
    if env_spec is None:
        env_spec = env.init_env_spec()
    env = env(stim_data, env_spec)
    return env


env_dict = {
    'ShapeNet': ShapeNetEnv,
}



import os
import re
from typing import Tuple, Dict, List
import random

import pandas as pd
import numpy as np
from numpy.typing import NDArray

from iwisdm.core import Stimulus
from iwisdm.envs.registration import Constant, EnvSpec, StimData
from iwisdm.utils.read_write import read_img


def compare_when(when_list):
    """
    Compare the when_list to get number of frames the task can take
    @param when_list: a list of "last%d"%k
    @return: the number of frames the task can take (max k)
    """
    # note, n_frames = compare_when + 1; if when_list is ['last0'], then there should be 1 frame
    return max(list(map(lambda x: get_k(x), when_list)))


def find_delays(when_list):
    """
    Find the delay frames in the when_list
    @param when_list: a list of "last%d"
    @return: the delays in the when_list
    """
    whens = sorted(list(map(lambda x: get_k(x), when_list)))
    assert len(set(whens)) == len(whens)
    no_delay = set(range(min(whens), max(whens)))
    delays = no_delay - set(whens)
    return list(map(lambda x: f'last_{x}', sorted(delays)))


def get_k(last_k: str):
    """
    Get the integer k from the string "last%d"%k
    @param last_k: last_k string
    @return: integer k in last_k
    """
    k_s = list(map(int, re.findall(r'\d+', last_k)))
    assert len(k_s) == 1
    return k_s[0]


def get_target_value(t):
    """
    Convert boolean target value to string for task answers
    @param t: target value
    @return: the task answer format of the target value
    """
    # Convert target t to string and convert True/False target values
    # to lower case strings for consistency with other uses of true/false
    # in vocabularies.
    t = str(t)
    if t is True or t == 'True':
        return 'true'
    if t is False or t == 'False':
        return 'false'
    return t


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


class SNConst(Constant):
    """
    singleton class, constants for the ShapeNet environment
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.INVALID = 'invalid'
        # define all available operators for constructing graphs
        self.LOGIC_OPS = ['And', 'Or', 'Xor', 'IsSame', "NotSame"]
        self.BOOL_OP = ['IsSame', 'Exist', 'And', 'Or', 'Xor', 'NotSame']
        self.ALLSPACES = ['left', 'right', 'top', 'bottom']

        # attributes available for given dataset (shapenet)
        # need to be updated if other datasets are used
        self.ATTRS = ['object', 'view_angle', 'category', 'location']


class SNEnvSpec(EnvSpec):
    """
    Environment specification for ShapeNet
    """

    def __init__(
            self,
            grid_size: Tuple[int, int] = (2, 2),
            max_delay: int = 2,
            delay_prob: float = 0.5,
            auto_gen_config: Dict = None,
            add_fixation_cue: True = False,
            canvas_size: int = 224,
            **kwargs
    ):
        if auto_gen_config is None:
            op_dict = {
                "Select":
                    {
                        "n_downstream": 4,
                        "downstream": ["GetLoc", "GetCategory", "GetObject"],
                        "same_children_op": False,
                        "min_depth": 1,
                        "min_op": 1,
                    },
                "GetCategory":
                    {
                        "n_downstream": 1,
                        "downstream": ["Select"],
                        "min_depth": 2,
                        "min_op": 2,
                    },
                "GetLoc":
                    {
                        "n_downstream": 1,
                        "downstream": ["Select"],
                        "min_depth": 2,
                        "min_op": 2,
                    },
                "GetObject":
                    {
                        "n_downstream": 1,
                        "downstream": ["Select"],
                        "min_depth": 2,
                        "min_op": 2,
                    },
                "IsSame":
                    {
                        "n_downstream": 2,
                        "downstream": ["GetLoc", "GetCategory", "GetObject"],
                        "sample_dist": [1 / 3, 1 / 3, 1 / 3],
                        "same_children_op": True,
                        "min_depth": 3,
                        "min_op": 7,
                    },
                "NotSame":
                    {
                        "n_downstream": 2,
                        "downstream": ["GetLoc", "GetCategory", "GetObject"],
                        "sample_dist": [1 / 3, 1 / 3, 1 / 3],
                        "same_children_op": True,
                        "min_depth": 3,
                        "min_op": 7,
                    },
                "And":
                    {
                        "n_downstream": 2,
                        "downstream": ["IsSame", "NotSame", "And", "Or"],
                        "same_children_op": False,
                        "min_depth": 4,
                        "min_op": 15,
                    },
                "Or":
                    {
                        "n_downstream": 2,
                        "downstream": ["IsSame", "NotSame", "And", "Or"],
                        "same_children_op": False,
                        "min_depth": 4,
                        "min_op": 15,
                    },
                "CONST":
                    {
                        "n_downstream": 0,
                        "downstream": [],
                        "sample_dist": [],
                        "same_children_op": False,
                        "min_depth": 1,
                        "min_op": 1,
                    }
            }
            auto_gen_config = {
                'op_dict': op_dict,
                # root_ops are the operators to begin a task
                'root_ops': ["IsSame", "And", "Or", "NotSame", "GetLoc", "GetCategory"],
                'boolean_ops': ["IsSame", "And", "Or", "NotSame", ],
                # all tasks end with select
                'leaf_op': ["Select"],
                'mid_op': ["Switch"],
                'max_op': 20,
                'max_depth': 10,
                'max_switch': 1,
                'switch_threshold': 0,
                'select_limit': False,
                'compare_const_prob': 1 / 15,
                'const_parent_ops': ["IsSame", "NotSame"],
                'indexable_get_ops': ["GetLoc", "GetCategory"],
            }
        self.add_fixation_cue = add_fixation_cue
        self.canvas_size = canvas_size
        super().__init__(
            grid_size,
            max_delay,
            delay_prob,
            auto_gen_config,
            **kwargs,
        )

    def sample_when(self, n: int = 1, existing_whens=None) -> list:
        """
        sample n 'lastk' values,
        avoid sampling multiple stimuli per frame
        @param n: how many values to sample
        @param existing_whens: list of existing 'lastk' values
        @return: list of n 'lastk' values
        """
        whens = list()
        i, count, n_delays = 0, 0, 0
        if existing_whens:
            max_k = compare_when(existing_whens)
            n_delays += len(find_delays(existing_whens))
            i += max_k
        else:
            whens.append(f'last{i}')
            count += 1

        while count < n:
            add_delay = np.random.random() < self.delay_prob
            i += 1
            if add_delay and n_delays < self.MAX_DELAY:  # delay, don't add lasti
                n_delays += 1
            else:
                count += 1
                whens.append(f'last{i}')
        return whens

    def check_whens(self, whens: List[str], existing_whens: List[str] = None):
        """
        Check if the whens are valid, i.e. only 1 stimulus per frame
        @param whens: the list of 'lastk' values
        @param existing_whens: existing list of 'lastk' values
        @return: resampled whens
        """
        # added check_whens to ensure 1 stimulus per frame
        existing_whens = set() if not existing_whens else set(existing_whens)
        len_ew = len(existing_whens)

        while len(set(whens) | existing_whens) != (len(whens) + len_ew):
            whens = self.sample_when(len(whens), existing_whens)
        return whens


class SNStimData(StimData):
    def __init__(
            self,
            dir_path: str = None,
            find_subdir: bool = True,
            df: pd.DataFrame = None,
            img_folder_path: str = None,
            splits: Dict = None
    ):
        super().__init__(dir_path, find_subdir, df, img_folder_path, splits)
        if self.find_subdir:
            for k, v in self.splits.items():
                if v:
                    assert v['path'] and v['df'] is not None, 'missing path or data for split {k}'
                    v['data'] = SNStimData(
                        dir_path=v['path'],
                        find_subdir=False,
                        df=v['df'],
                        img_folder_path=v['path'],
                        splits=None
                    )
        self.ATTR_DICT = self.get_all_attributes(self.df)
        self.attr_with_mapping = self.get_attr_str_mapping()
        self.ALLCATEGORIES = list(self.ATTR_DICT.keys())
        self.ALLOBJECTS = {c: list(self.ATTR_DICT[c].keys()) for c in self.ATTR_DICT}
        self.ALLVIEWANGLES = {c: {obj: list(info.keys()) for obj, info in d.items()} for c, d in self.ATTR_DICT.items()}
        self.ALL_ATTRS = {'location', 'category'}

    def get_object(self, obj: Stimulus, obj_size: Tuple[int, int], mode: str = None) -> NDArray:
        """
        Get the image array of the object from the provided data directory
        @param obj: a stimulus instance with attribute values
        @param obj_size: the size of the stimulus on the canvas
        @param mode: the split, [train, test, validation]
        @return: image array, RGB format
        """
        if mode:
            image_path = self.splits[mode]['path']
            if self.splits[mode]['data']:
                attr_dict = self.splits[mode]['data'].ATTR_DICT
            else:
                attr_dict = self.get_all_attributes(self.splits[mode]['df'])
        else:
            image_path = self.img_folder_path
            attr_dict = self.ATTR_DICT

        refs = attr_dict[obj.category][obj.object][obj.view_angle]
        if not refs:
            raise ValueError(f'ShapeNet object with '
                             f'category {obj.category}, identity {obj.object}, view angle {obj.view_angle} not found')

        obj_ref = random.choice(refs)
        obj_path = os.path.join(image_path, f'{obj_ref}/image.png')

        return read_img(obj_path, obj_size, color_format='RGB')

    @staticmethod
    def get_all_attributes(df):
        """
        Get all shapenet attributes from the dataset, including the hierarchy
        @return: dictionary of the format {category: {object: [view angles]}}
        """
        ATTR_DICT = dict()

        for cat in df['ctg_mod'].unique():
            ATTR_DICT[cat] = dict()
            unique_obj = df[df['ctg_mod'] == cat]['obj_mod'].unique()
            for obj in unique_obj:
                ATTR_DICT[cat][obj] = dict()
                unique_va = df[(df['ctg_mod'] == cat) & (df['obj_mod'] == obj)]['ang_mod'].unique()
                for va in unique_va:
                    refs = df[
                        (df['ctg_mod'] == cat) & (df['obj_mod'] == obj) & (df['ang_mod'] == va)
                        ]['ref'].unique()
                    ATTR_DICT[cat][obj][va] = list(map(int, refs))
        return ATTR_DICT

    def get_attr_str_mapping(self):
        """
        get the mapping from integer attribute value to string description
        e.g. 0 -> 'airplane', 1 -> 'car'
        @return: the mapping dictionary of the format {attribute: {int: str}}
        """
        self.attr_with_mapping = dict()
        if 'ctg' in self.df.columns.values:
            IDX2Category = dict()
            for cat, _ in self.ATTR_DICT.items():
                labels = self.df[self.df['ctg_mod'] == cat]['ctg']
                assert len(set(labels)) == 1, f'found more than 1 label for the cateogry {cat}'
                IDX2Category[cat] = set(labels).pop()
            self.attr_with_mapping['category'] = IDX2Category
        return self.attr_with_mapping


DATA = SNConst()



""" procedural task generation.

based on the specified connectivity between operators,
construct random task graphs by adding operators layer by layer
also constructs tasks based on the generated graphs
"""

from collections import defaultdict

import networkx as nx
from iwisdm.core import Operator, Attribute, Task
from iwisdm.envs.registration import EnvSpec
from typing import Tuple, Union

# Tuples of the graph object, the root root_op number, and the number of operators,
# needed to compose graphs in switch generation
GRAPH_TUPLE = Tuple[nx.DiGraph, int, int]
TASK = Tuple[Union[Operator, Attribute], Task]


class TaskGenerator:
    """
    class for generating random task graphs and tasks based on the specified operator connectivity
    """

    def __init__(self, env_spec: EnvSpec):
        self.env_spec = env_spec
        # dictionary specifying which operators can follow an operator,
        # e.g. In ShapeNet, GetCategory follows selecting an object
        # 4 operators are related to Select: category, location, view_angle, and the exact object
        # if the operator is None, then a random constant is sampled for that attribute
        self.config = env_spec.auto_gen_config

        self.op_dict = self.config['op_dict']
        self.root_ops = self.config['root_ops']
        self.boolean_ops = self.config['boolean_ops']
        op_dict = defaultdict(dict, **self.op_dict)
        self.op_depth_limit = {k: v['min_depth'] for k, v in op_dict.items()}
        self.op_operators_limit = {k: v['min_op'] for k, v in op_dict.items()}

    def generate_task(self, *args):
        raise NotImplementedError



# code based on https://github.com/google/cog

"""A bank of available tasks."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict
import random

from iwisdm.envs.shapenet.task_generator import TemporalTask
from iwisdm.envs.shapenet import task_generator as tg
import iwisdm.envs.shapenet.stim_generator as sg
import iwisdm.envs.shapenet.registration as env_reg


class ExistCategoryOfTemporal(TemporalTask):
    """Check if  object on given frame has the same category with the object on another frame."""

    def __init__(self, whens, first_shareable=None):
        super(ExistCategoryOfTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        category1 = tg.GetCategory(objs1)
        objs2 = tg.Select(category=category1, when=when2)
        self._operator = tg.Exist(objs2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class ExistViewAngleOfTemporal(TemporalTask):
    """Check if  object on given frame has the same view angle with the object on another frame."""

    def __init__(self, whens, first_shareable=None):
        super(ExistViewAngleOfTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = whens[0], whens[1]

        objs1 = tg.Select(when=when1)
        view_angle = tg.GetViewAngle(objs1)
        objs2 = tg.Select(view_angle=view_angle, when=when2)
        self._operator = tg.Exist(objs2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class ExistObjectOfTemporal(TemporalTask):
    """Check if on given frame has the same object with another frame"""

    def __init__(self, whens, first_shareable=None):
        super(ExistObjectOfTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        obj = tg.GetObject(objs1)
        objs2 = tg.Select(object=obj, when=when2)
        self._operator = tg.Exist(objs2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class CompareCategoryTemporal(TemporalTask):
    """Compare objects on chosen frames are of the same category or not."""

    def __init__(self, whens, first_shareable=None):
        super(CompareCategoryTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        objs2 = tg.Select(when=when2)
        a1 = tg.GetCategory(objs1)
        a2 = tg.GetCategory(objs2)
        self._operator = tg.IsSame(a1, a2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class CompareViewAngleTemporal(TemporalTask):
    """Compare objects on chosen frames are of the same view angle or not"""

    def __init__(self, whens, first_shareable=None):
        super(CompareViewAngleTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        objs2 = tg.Select(when=when2)
        a1 = tg.GetViewAngle(objs1)
        a2 = tg.GetViewAngle(objs2)
        self._operator = tg.IsSame(a1, a2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class CompareLocTemporal(TemporalTask):
    """Compare objects on chosen frames are of the same location or not"""

    def __init__(self, whens, first_shareable=None):
        super(CompareLocTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        objs2 = tg.Select(when=when2)
        a1 = tg.GetLoc(objs1)
        a2 = tg.GetLoc(objs2)
        self._operator = tg.IsSame(a1, a2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class CompareObjectTemporal(TemporalTask):
    """Compare objects on chosen frames are of the same identity or not"""

    def __init__(self, whens, first_shareable=None):
        super(CompareObjectTemporal, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        objs1 = tg.Select(when=when1)
        objs2 = tg.Select(when=when2)
        a1 = tg.GetObject(objs1)
        a2 = tg.GetObject(objs2)
        self._operator = tg.IsSame(a1, a2)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


class SequentialCategoryMatch(TemporalTask):
    # nback category
    def __init__(self, whens, first_shareable=None, n_compare=1):
        super(SequentialCategoryMatch, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]
        max_k = env_reg.compare_when([when1, when2])
        if n_compare * 2 + 1 > max_k:
            n_compare = max_k // 2
        total_frames = n_compare * 2 + random.randint(0, max_k - (n_compare * 2) + 1)

        sample_objs = [tg.Select(when=f'last{total_frames - i - 1}') for i in range(n_compare)]
        response_objs = [tg.Select(when=f'last{i}') for i in range(n_compare)]
        sample_cats = [tg.GetCategory(obj) for obj in sample_objs]
        response_cats = [tg.GetCategory(obj) for obj in response_objs]
        is_sames = [tg.IsSame(sample, response) for sample, response in zip(sample_cats, response_cats)]
        if n_compare == 1:
            self._operator = is_sames[0]
        else:
            ands = tg.And(is_sames[0], is_sames[1])
            for is_same1, is_same2 in zip(is_sames[2::2], is_sames[3::2]):
                ands = tg.And(tg.And(is_same1, is_same2), ands)
            self._operator = ands
        self.n_frames = total_frames


class DelayedCDM(TemporalTask):
    # contextual decision making

    def __init__(self, whens=None, first_shareable=None, attrs=None):
        super(DelayedCDM, self).__init__(whens=whens, first_shareable=first_shareable)
        when1, when2 = self.whens[0], self.whens[1]

        objs1 = tg.Select(when=when1)
        objs2 = tg.Select(when=when2)

        if attrs:
            assert len(attrs) == 3
            attr0 = attrs[0]
            attrs = attrs[1::]
        else:
            attr0 = random.choice(env_reg.DATA.ATTRS)
            attrs = random.sample(env_reg.DATA.ATTRS, 2)
        const_attr0 = sg.random_attr(attr0)
        const_attrs = [sg.random_attr(attr) for attr in attrs]
        condition = tg.IsSame(const_attr0, tg.get_family_dict[attr0](objs1))
        do_if = tg.IsSame(const_attrs[0], tg.get_family_dict[attrs[0]](objs2))
        do_else = tg.IsSame(const_attrs[1], tg.get_family_dict[attrs[1]](objs2))
        self._operator = tg.Switch(condition, do_if, do_else, both_options_avail=False)
        self.n_frames = env_reg.compare_when([when1, when2]) + 1


task_family_dict = OrderedDict([
    ('ExistCategoryOf', ExistCategoryOfTemporal),
    ('ExistViewAngleOf', ExistViewAngleOfTemporal),
    ('ExistObjectOf', ExistObjectOfTemporal),
    ('CompareViewAngle', CompareViewAngleTemporal),
    ('CompareCategory', CompareCategoryTemporal),
    ('CompareObject', CompareObjectTemporal),
    ('CompareLoc', CompareLocTemporal),
    ('SequentialCategory', SequentialCategoryMatch),
    ('DelayedCDM', DelayedCDM),
])



""" procedural task generation.

based on the specified connectivity between operators,
construct random task graphs by adding operators layer by layer
also constructs tasks based on the generated graphs
"""

import random
from typing import Tuple, Union, List

import numpy as np
import networkx as nx

from iwisdm.core import Operator, Attribute, Task
from iwisdm.utils.auto_task_gen import TaskGenerator
import iwisdm.envs.shapenet.task_generator as tg
from iwisdm.envs.shapenet.registration import SNEnvSpec

# Tuples of the graph object, the root root_op number, and the number of operators,
# needed to compose graphs in switch generation
GRAPH_TUPLE = Tuple[nx.DiGraph, int, int]
TASK = Tuple[Union[Operator, Attribute], Task]


class SNTaskGenerator(TaskGenerator):
    """
    class for generating random task graphs and tasks based on the specified operator connectivity
    """

    def __init__(self, env_spec: SNEnvSpec):
        """
        constructor for the SNTaskGenerator
        @param env_spec: the task generation configuration,
        see iwisdm/envs/shapenet/registration.py for default env_spec
        """
        super().__init__(env_spec)
        self.boolean_ops = self.config['boolean_ops']
        self.max_op = self.config['max_op']
        self.max_depth = self.config['max_depth']
        self.select_limit = self.config['select_limit']
        self.max_switch = self.config['max_switch']
        self.switch_threshold = self.config['switch_threshold']
        self.compare_const_prob = self.config['compare_const_prob']
        self.const_parent_ops = self.config['const_parent_ops']
        self.indexable_get_ops = self.config['indexable_get_ops']

    @staticmethod
    def count_depth_and_op(op: Operator) -> Tuple[int, int]:
        """
        count the number of operators and the depth of the operator graph
        @param op: the root opeartor with nested sucessors
        @return: [node count, depth count]
        """
        op_count, depth_count = 0, 0
        if not isinstance(op, Operator):
            return op_count, depth_count
        if op.child:
            # iterate over the nodes, and increment count
            depth_counts = list()
            for c in op.child:
                add_count = SNTaskGenerator.count_depth_and_op(c)
                op_count += add_count[0]
                depth_counts += [add_count[1]]
            depth_count += max(depth_counts)
        else:
            # at the leaf nodes
            return 1, 1
        op_count += 1
        depth_count += 1
        return op_count, depth_count

    @staticmethod
    def switch_generator(conditional: GRAPH_TUPLE, do_if: GRAPH_TUPLE, do_else: GRAPH_TUPLE) -> GRAPH_TUPLE:
        """
        function to generate a switch operator based on the conditional, do_if, and do_else subtask graphs
        stitches the 3 subtask graphs together to form a switch task graph
        @param conditional:
        @param do_if:
        @param do_else:
        @return:
        """
        # combine the 3 subtasks graphs into the switch task graph by using networkx compose_all
        do_if_graph, do_if_root, do_if_node = do_if
        do_else_graph, do_else_root, do_else_node = do_else
        conditional_graph, conditional_root, conditional_node = conditional

        # combine all 3 graphs and add edges
        G: nx.DiGraph = nx.compose_all([do_if_graph, do_else_graph, conditional_graph])
        switch_count = conditional_node + 1
        G.add_node(switch_count, label='Switch')
        # note that all directed edges are reversed later at task construction
        # after reversal, switch operator is connected to the selects of do_if and do_else
        # the root of conditional is then connected to the switch operator
        G.add_edge(do_if_node, switch_count)
        G.add_edge(do_else_node, switch_count)
        G.add_edge(switch_count, conditional_root)
        return G, switch_count, switch_count

    def sample_root_helper(self, max_op, max_depth) -> Operator:
        """
        sample a root operator given constraints on the task complexity
        @param max_op: maximum number of operators allowed
        @param max_depth: maximum depth of the task graph
        @return: the root operator that can complete a subtask graph
        """
        op_depth_limit = self.op_depth_limit
        op_operators_limit = self.op_operators_limit

        depth_filter = [op for op, v in op_depth_limit.items() if (v <= max_depth) and (op in self.root_ops)]
        both_filter = [op for op in depth_filter if op_operators_limit[op] <= max_op]
        if not both_filter:
            depth = {op: v for op, v, in op_depth_limit.items() if op in self.root_ops}
            root_op = {op: v for op, v, in op_operators_limit.items() if op in self.root_ops}
            raise ValueError(f'The specified task complexity is too low given the available root operators\n'
                             f'The minimum depth of the specified root operators is: {depth}\n'
                             f'The minimum depth of the specified root operators is: {root_op}')
        return np.random.choice(both_filter)

    def sample_children_helper(self, op_name, op_count, max_op, cur_depth, max_depth):
        """
        helper function to ensure the task complexity is bounded, and return the sampled child operator
        max_depth is a tighter bound than max_op since operators have to complete their subtask graph
        :param op_name: the current operator
        :param op_count: the current number of operators
        :param max_op: the maximum number of operators allowed
        :param cur_depth: the current depth
        :param max_depth: the maximum depth of the task graph
        :return: a randomly sampled operator to follow the parent node
        """
        op_depth_limit = self.op_depth_limit
        op_operators_limit = self.op_operators_limit
        downstream_ops = self.op_dict[op_name]["downstream"]

        min_add_depth_filter = {op: (max_depth - (cur_depth + self.op_depth_limit[op] - 1)) for op in downstream_ops if
                                (cur_depth + op_depth_limit[op] <= max_depth)}
        min_add_op_filter = {op: (max_op - (op_count + op_operators_limit[op])) for op in downstream_ops if
                             (op_count + op_operators_limit[op] <= max_op)}

        complexity_downstream = [op_depth_limit[op] for op in downstream_ops]
        min_complexity_downstream = [op for op in downstream_ops if op_depth_limit[op] == min(complexity_downstream)]
        if "sample_dist" in self.op_dict[op_name]:  # sample dist overrides complexity bounding
            return np.random.choice(self.op_dict[op_name]["downstream"], p=self.op_dict[op_name]["sample_dist"])
        if not min_add_depth_filter:
            return np.random.choice(min_complexity_downstream)

        if max(min_add_depth_filter.values()) > 0:
            # if added operator sub-graph can fit and have left over depth
            filtered_ops = [op for op, diff in min_add_depth_filter.items() if diff > 0]
            both_filter = {k: v for k, v in min_add_op_filter.items() if k in filtered_ops}
            if both_filter:
                if max(both_filter.values()) > 0:
                    both_filter = [k for k, v in both_filter.items() if v > 0]
                    return np.random.choice(both_filter)
                elif max(both_filter.values()) == 0:
                    both_filter = [k for k, v in both_filter.items() if v == 0]
                    return np.random.choice(both_filter)
            return np.random.choice(min_complexity_downstream)
        elif max(min_add_depth_filter.values()) == 0:
            # if only certain operator sub-graphs can fit
            filtered_ops = [op for op, diff in min_add_depth_filter.items() if diff == 0]
            both_filter = {k: v for k, v in min_add_op_filter.items() if k in filtered_ops}
            if both_filter:
                if max(both_filter.values()) >= 0:
                    both_filter = [k for k, v in both_filter.items() if v == max(both_filter.values())]
                    return np.random.choice(both_filter)
            # cannot do anything about bounding, just complete the minimum task graph
            return np.random.choice(min_complexity_downstream)
        else:
            # cannot do anything about bounding, just complete the minimum task graph
            return np.random.choice(min_complexity_downstream)

    def sample_children_op(
            self,
            op_name: str,
            op_count: int,
            max_op: int,
            cur_depth: int,
            max_depth: int,
            select_op: bool,
            select_downstream: List[str]
    ) -> List[str]:
        """
        sample the children operators given the operator name
        :param op_name: the name of the parent node
        :param op_count: operator count
        :param max_op: max number of operators allowed
        :param cur_depth: current depth
        :param max_depth: max depth allowed for the task graph
        :param select_op: Boolean, does select follow an operator
        :param select_downstream: the downstream options that can be sampled from
        :return: list of operators
        """

        n_downstream = self.op_dict[op_name]["n_downstream"]  # how many children need to be sampled
        if n_downstream == 1:
            return [self.sample_children_helper(op_name, op_count, max_op, cur_depth, max_depth)]
        elif op_name == 'Select':
            children = list()  # append children operators

            if cur_depth + 2 > max_depth or op_count + 2 > max_op:
                return ['None' for _ in range(n_downstream)]
            else:
                op_sampled = False
                if select_op:  # if Select is followed by a Get operator
                    if select_downstream is None:
                        select_downstream = self.op_dict['Select']['downstream']
                    for _ in range(n_downstream):
                        if np.random.random() < 0.8:  # make sure not too many operators follow select
                            children.append('None')
                        else:
                            if select_downstream and not op_sampled:  # if the list is not empty, sample a downstream op
                                get = random.choice(select_downstream)
                                children.append(get)
                                op_sampled = True
                                select_downstream.remove(get)
                            else:
                                children.append('None')
                    return children
                else:
                    return ['None' for _ in range(n_downstream)]
        else:
            if self.op_dict[op_name]["same_children_op"]:
                child = self.sample_children_helper(op_name, op_count, max_op, cur_depth, max_depth)
                return [child for _ in range(n_downstream)]
            else:
                return [self.sample_children_helper(op_name, op_count, max_op, cur_depth, max_depth) for _ in
                        range(n_downstream)]

    def branch_generator(
            self,
            G: nx.DiGraph,
            root_op: str,
            op_count: int,
            max_op: int,
            cur_depth: int,
            max_depth: int,
            select_op=False,
            select_downstream=None,
    ) -> int:
        """
        modifies G in place
        function to complete a branch of subtask graph based on a root operator,
        @param G:
        @param root_op:
        @param op_count:
        @param max_op:
        @param cur_depth:
        @param max_depth:
        @param select_op:
        @param select_downstream:
        @return: the current operator count
        """
        # function to complete a branch of subtask graph based on a root operator,
        # modifies G in place, and returns the leaf node number
        if root_op == 'None':
            return None  # don't add any nodes or edges
        elif root_op == 'CONST':
            return op_count  # don't add any nodes or edges
        else:
            # Exist operator always follows a select with operator as child
            select_op = True if root_op == 'Exist' else select_op

            children = self.sample_children_op(
                op_name=root_op,
                op_count=op_count + 1,
                max_op=max_op,
                cur_depth=cur_depth + 1,
                max_depth=max_depth,
                select_op=select_op,
                select_downstream=select_downstream
            )
            cur_depth += 1  # increment cur_depth count

            if root_op == 'Select' and any('Get' in c for c in children):
                # stop generating overly complex tasks where selects with operator children follow more operators
                select_downstream = ['None'] * 4
                select_op = False

            parent = op_count
            compare_const_prob, const_parent_ops = self.compare_const_prob, self.const_parent_ops
            if root_op in const_parent_ops:
                # current environment only allows comparing category and location const
                # do not compare constants if comparing view_angle and object
                # remove this check if there's mapping for object/view_angle idx
                if not any(op not in self.indexable_get_ops for op in children):
                    if np.random.random() < compare_const_prob:
                        # if the random number is less than the probability
                        # then change one of the indexable Get operators to CONST
                        idx = random.choice(range(len(children)))
                        children[idx] = 'CONST'
                else:
                    if compare_const_prob == 1.0:
                        # force compare to CONST by changing children to indexable Get operators
                        if self.op_dict[root_op]['same_children_op']:
                            n_downstream = self.op_dict[root_op]["n_downstream"]
                            child = random.choice(self.indexable_get_ops)
                            children = [child for _ in range(n_downstream)]
                        else:
                            children = [
                                random.choice(self.indexable_get_ops)
                                if node not in self.indexable_get_ops else node
                                for node in children
                            ]
                        idx = random.choice(range(len(children)))
                        children[idx] = 'CONST'
            for op in children:  # loop over sampled children and modify the graph in place
                if op != 'None':  # if the operator is an operator
                    child = op_count + 1
                    # recursively generate branches based on the child operator
                    # op_count is incremented based on how many nodes were added in the child branch call
                    op_count = self.branch_generator(
                        G, op, child,
                        max_op, cur_depth, max_depth,
                        select_op, select_downstream
                    )
                    G.add_node(child, label=op)  # modify the graph
                    G.add_edge(parent, child)
            return op_count

    def subtask_graph_generator(
            self,
            count=0,
            max_op=20,
            max_depth=10,
            select_limit=False,
            root_op=None,
    ) -> GRAPH_TUPLE:
        """
        function for generating subtask graphs
        uses networkx to compose the task graphs
        @param count: the root_op number of the root
        @param max_op: the maximum number of operators allowed in the task graph
        @param max_depth: the maximum depth of the task graph
        @param select_limit: whether to add operator after selects. if True, then constants are sampled.
        If false, then operators could be sampled
        @param root_op: the root operator

        @return: GRAPH_TUPLE of (nx.DiGraph(), node number of the graph root, operator count)
        """
        # initialize the graph and save the root_op number of the root
        G = nx.DiGraph()
        root = count

        op_count = count

        root_op = root_op if root_op else self.sample_root_helper(max_op, max_depth)
        G.add_node(op_count, label=root_op)

        select_downstream = ['None'] * 4 if select_limit else None

        op_count = self.branch_generator(
            G, root_op, op_count,
            max_op, 0, max_depth,
            select_downstream=select_downstream
        )
        return G, root, op_count

    def generate_task(self) -> Tuple[GRAPH_TUPLE, TASK]:
        """
        function to generate a random task graph and corresponding task
        :param max_switch: the maximum number of switch operators allowed
        :param switch_threshold: float in [0,1], how likely a switch operator is sampled, higher is more likely
        :param max_op: max number of operators allowed
        :param max_depth: max depth of the task graph
        :param select_limit: whether to add operator after selects. if True, then constants are sampled.
        If false, then operators could be sampled
        :return: The random task graph tuple and task tuple
        """
        count = 0
        whens = dict()

        # generate a subtask graph and the actual task
        subtask_graph = self.subtask_graph_generator(
            count=count,
            max_op=self.max_op,
            max_depth=self.max_depth,
            select_limit=self.select_limit,

        )
        subtask, whens = tg.subtask_generation(self.env_spec, subtask_graph, existing_whens=whens)
        count = subtask_graph[2] + 1  # start a new subtask graph node number according to old graph
        for _ in range(self.max_switch):
            if random.random() < self.switch_threshold:  # if add switch
                new_task_graph = self.subtask_graph_generator(
                    count=count,
                    max_op=self.max_op,
                    max_depth=self.max_depth,
                    select_limit=self.select_limit
                )
                count = new_task_graph[2] + 1
                conditional = self.subtask_graph_generator(
                    count=count,
                    max_op=self.max_op,
                    max_depth=self.max_depth,
                    select_limit=self.select_limit,
                    root_op=random.choice(self.boolean_ops)
                )
                conditional_task, whens = tg.subtask_generation(self.env_spec, conditional, existing_whens=whens)
                if random.random() < 0.5:  # randomly split the do_if and do_else tasks
                    do_if = subtask_graph
                    do_if_task = subtask
                    do_else = new_task_graph
                    do_else_task, whens = tg.subtask_generation(self.env_spec, do_else, existing_whens=whens)
                else:
                    do_if = new_task_graph
                    do_if_task, whens = tg.subtask_generation(self.env_spec, do_if, existing_whens=whens)
                    do_else = subtask_graph
                    do_else_task = subtask
                subtask_graph = self.switch_generator(conditional, do_if, do_else)
                count = subtask_graph[2] + 1
                subtask = tg.switch_generation(conditional_task, do_if_task, do_else_task, whens)
        return subtask_graph, subtask



"""
Code based on 'A Dataset and Architecture for Visual Reasoning with a Working Memory', Guangyu Robert Yang, et al.
Paper: https://arxiv.org/abs/1803.06092
Code: https://github.com/google/cog

High-level API for generating stimuli.

Objects are first generated abstractly, with high-level specifications
like location='random'.
Abstract relationships between objects can also be specified.

All objects and relationships are then collected into a ObjectSet.
The ObjectSet object can interpret the abstract specifications and instantiate
the stimuli in each trial.

Rendering function generates movies based on the instantiated stimuli
"""

import json
from collections import defaultdict, OrderedDict
from typing import Tuple, Union, Dict, Callable, List

import networkx as nx
import numpy as np
import random

from iwisdm.core import (
    Task,
    Operator
)
from iwisdm.envs.shapenet.registration import SNStimData, SNEnvSpec
import iwisdm.envs.shapenet.stim_generator as sg
import iwisdm.envs.shapenet.registration as env_reg


def obj_str(
        location=None,
        obj=None,
        category=None,
        view_angle=None,
        when=None
):
    """Get a string describing an object with attributes."""

    location = location or sg.Location(space=sg.random_grid_space())
    category = category or sg.SNCategory(None)
    obj = obj or sg.SNObject(category=category, value=None)
    view_angle = view_angle or sg.SNViewAngle(sn_object=obj, value=None)

    sentence = []
    if when is not None:
        sentence.append(when)
    if isinstance(category, sg.SNAttribute) and category.has_value():
        sentence.append(str(category))
    if isinstance(view_angle, sg.SNAttribute) and view_angle.has_value():
        sentence.append(str(view_angle))
    if isinstance(obj, sg.SNAttribute) and obj.has_value():
        sentence.append(str(obj))
    if isinstance(location, sg.SNAttribute) and location.has_value():
        sentence.append(str(location))
    else:
        sentence.append('object')

    if isinstance(category, Operator):
        sentence += ['with', str(category)]
    if isinstance(view_angle, Operator):
        sentence += ['with', str(view_angle)]
    if isinstance(location, Operator):
        sentence += ['at', str(location)]
    if isinstance(obj, Operator):
        sentence += ['with', str(category)]
    return ' '.join(sentence)


class Skip(object):
    """Skip this operator."""

    def __init__(self):
        pass


class SNOperator(Operator):
    constants = env_reg.DATA
    _stim_data: SNStimData = None
    env_spec: SNEnvSpec = None

    @property
    def stim_data(self):
        return self._stim_data

    @stim_data.setter
    def stim_data(self, value):
        self._stim_data = value


class Select(SNOperator):
    """Selecting the objects that satisfy properties."""

    def __init__(self,
                 location=None,
                 category=None,
                 object=None,
                 view_angle=None,
                 attrs=None,
                 when=None,
                 space_type=None):
        super(Select, self).__init__()

        if attrs:
            for attr in attrs:
                if isinstance(attr, sg.Location):
                    location = attr
                elif isinstance(attr, sg.SNCategory):
                    category = attr
                elif isinstance(attr, sg.SNObject):
                    object = attr
                elif isinstance(attr, sg.SNViewAngle):
                    view_angle = attr
        location = location or sg.Location(space=sg.random_attr('space'), value=None)
        category = category or sg.SNCategory(None)
        object = object or sg.SNObject(category, None)
        view_angle = view_angle or sg.SNViewAngle(object, None)

        self.location, self.category, self.object, self.view_angle = location, category, object, view_angle
        self.set_child([location, category, object, view_angle])

        self.when = when
        self.space_type = space_type

    def __str__(self):
        return obj_str(
            self.location,
            self.category,
            self.object,
            self.view_angle,
            self.when
        )

    def __call__(self, objset: sg.ObjectSet, epoch_now: int):
        """Return subset of objset."""
        location = self.location(objset, epoch_now)
        category = self.category(objset, epoch_now)
        object = self.object(objset, epoch_now)
        view_angle = self.view_angle(object, epoch_now)

        if env_reg.DATA.INVALID in (location, category, object, view_angle):
            return env_reg.DATA.INVALID

        if self.space_type is not None:
            space = location.get_space_to(self.space_type)
        else:
            space = sg.Space()

        subset = objset.select(
            epoch_now,
            space,
            category,
            object,
            view_angle,
            self.when,
        )
        return subset

    def __hash__(self):
        if self.child:
            c_s = sorted(self.child, key=lambda o: o.__class__.__name__)
            return hash(tuple(
                [self.__class__.__name__] +
                [c for c in c_s if isinstance(c, Operator)] +
                [self.when]
            ))
        else:
            return hash(self.__class__.__name__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if len(self.child) != len(other.child):
                return False
            else:
                for c, o_c in zip(
                        sorted(self.child, key=lambda o: o.__class__.__name__),
                        sorted(other.child, key=lambda o: o.__class__.__name__)
                ):
                    if isinstance(c, Operator) and isinstance(o_c, Operator):
                        if c != o_c:
                            return False
                if self.when != other.when:
                    return False
                return True
        return False

    def copy(self):
        return Select(location=self.location, category=self.category, object=self.object, view_angle=self.view_angle,
                      when=self.when, space_type=self.space_type)

    def hard_update(self, obj: sg.Object):
        """
        change the attributes based on the attributes of the provided obj
        :param obj: an object on a frame that select needs to corresponds to
        :return: True if modification is successful
        """
        assert obj.check_attrs()

        self.category = obj.category
        self.object = obj.object
        self.view_angle = obj.view_angle
        self.location = obj.location
        return True

    def soft_update(self, obj: sg.Object):
        """
        change the attributes based on the attributes of the provided obj
        don't change the attribute if no value has been assigned

        :param obj: an object on a frame that select needs to corresponds to
        :return: True if modification is successful
        """
        assert obj.check_attrs()

        self.category = obj.category if self.category.has_value() else self.category
        self.object = obj.object if self.object.has_value() else self.object
        self.view_angle = obj.view_angle if self.view_angle.has_value() else self.view_angle
        self.location = obj.location if self.location.has_value() else self.location
        return True

    def get_expected_input(self, should_be, objset, epoch_now):
        """Guess objset for Select operator.

        Optionally modify the objset based on the target output (should_be),
        and pass down the supposed input attributes

        There are two sets of attributes to compute:
        (1) The attributes of the expected input, i.e. the attributes to select
        (2) The attributes of new object being added to the objset

        There are two main scenarios:
        (1) the target output is empty
        (2) the target output is not empty, then it has to be a list with a single
        Object instance in it

        When (1) the target output is empty, the goal is to make sure
        attr_newobject and attr_expectedinput differ by a single attribute

        When (2) the target output is not empty, then
        attr_newobject and attr_expectedinput are the same

        We first decide the attributes of the expected inputs.

        If target output is empty, then expected inputs is determined solely
        based on its actual inputs. If the actual inputs can already be computed,
        then use it, and skip the traversal of following nodes. Otherwise, use a
        random allowed value.

        If target output is not empty, again, if the actual inputs can be computed,
        use it. If the actual input can not be computed, but the attribute is
        specified by the target output, use that. Otherwise, use a random value.

        Then we determine the new object being added.

        If target output is empty, then randomly change one of the attributes that
        is not None.
        For self.space attribute, if it is an operator, then add the object at the
        opposite side of space

        If target output is not empty, use the attributes of the expected input

        Args:
          should_be: a list of Object instances
          objset: objset instance
          epoch_now: int, the current epoch

        Returns:
          objset: objset instance
          space: supposed input
          category: supposed input
          object: supposed input
          view_angle: supposed input

        Raises:
          TypeError: when should_be is not a list of Object instances
          NotImplementedError: when should_be is a list and has length > 1
        """

        if should_be is None:
            raise NotImplementedError('This should not happen.')

        if should_be:
            # Making sure should_be is a length-1 list of Object instance
            for s in should_be:
                if not isinstance(s, sg.Object):
                    raise TypeError('Wrong type for items in should_be: '
                                    + str(type(s)))

            if len(should_be) > 1:
                for s in should_be:
                    print(s)
                raise NotImplementedError()

            obj_should_be = should_be[0]

            attr_new_object = list()
            attr_type_not_fixed = list()

            # First evaluate the inputs of the operator that can be evaluated
            for attr_type in ['location', 'category', 'object', 'view_angle']:
                a = getattr(self, attr_type)
                attr = a(objset, epoch_now)
                # If the input is successfully evaluated
                if attr is not env_reg.DATA.INVALID and attr.has_value():
                    if attr_type == 'location' and self.space_type is not None:
                        attr = attr.get_space_to(self.space_type)
                        print('space type is not none')
                    attr_new_object.append(attr)
                else:
                    attr_type_not_fixed.append(attr_type)

            # Add an object based on these attributes
            # Note that, objset.add() will implicitly select the objset
            obj: sg.Object = sg.Object(attr_new_object, when=self.when)
            obj = objset.add(obj, epoch_now, add_if_exist=False)

            if obj is None:
                return objset, Skip(), Skip(), Skip(), Skip()

            # If some attributes of the object are given by should_be, use them
            # change the attributes in obj based on should_be
            for attr_type in attr_type_not_fixed:
                a = getattr(obj_should_be, attr_type)
                if a.has_value():
                    if attr_type in ['category', 'object', 'view_angle']:
                        obj.change_attr(a)
                    else:
                        setattr(obj, attr_type, a)

            # If an attribute of select is an operator, then the expected input is the value of obj
            # attr_new_object is the output for the subsequent root children
            attr_expected_in = list()
            for attr_type in ['location', 'category', 'object', 'view_angle']:
                a = getattr(self, attr_type)
                if isinstance(a, Operator):
                    # Only operators need expected_in
                    # removed checking for location, since we're not flipping attributes
                    attr = getattr(obj, attr_type)
                    attr_expected_in.append(attr)
                else:
                    attr_expected_in.append(Skip())

        # no target output given, only happens when parent root is Exist, and Exist has no expected output
        # Basically if the target output is an empty set, then we place a different object.
        # We choose to place a different object here in order to prevent the network from solving some
        # tasks by simply counting the number of object. The object we
        # place differs from the object to be selected by only one attribute
        if not should_be:
            # First determine the attributes to flip later
            attr_type_to_flip = list()
            for attr_type in ['location', 'category', 'object', 'view_angle']:
                a = getattr(self, attr_type)
                # If attribute is operator or is a specified value
                if isinstance(a, Operator) or a.has_value():
                    attr_type_to_flip.append(attr_type)

            # Now generate expected input attributes
            attr_expected_in = list()
            attr_new_object = list()
            for attr_type in ['location', 'category', 'object', 'view_angle']:
                a = getattr(self, attr_type)
                attr = a(objset, epoch_now)
                if isinstance(a, Operator):
                    if attr is env_reg.DATA.INVALID:
                        # Can not be evaluated yet, then randomly choose one
                        attr = sg.random_attr(attr_type)
                    attr_expected_in.append(attr)
                else:
                    attr_expected_in.append(Skip())

                if attr_type in attr_type_to_flip:
                    # Candidate attribute values for the new object
                    attr_new_object.append(attr)

            # Randomly pick one attribute to flip
            attr_type = random.choice(attr_type_to_flip)
            i = attr_type_to_flip.index(attr_type)
            if attr_type == 'location':
                # if the location of the object is determined by the location another object
                # e.g. Exist->Select->GetLoc->Select:
                # check if exist object with location of a specified category/object/view_angle

                # If flip location, place it in the opposite direction
                location = attr_new_object[i]
                attr_new_object[i] = location.get_opposite_space_to(self.space_type)
            else:
                # Select a different attribute
                attr_new_object[i] = sg.another_attr(attr_new_object[i])
                # Not flipping location, so place it in the correct direction
                if 'location' in attr_type_to_flip:
                    j = attr_type_to_flip.index('location')
                    location = attr_new_object[j]
                    attr_new_object[j] = location.get_space_to(self.space_type)

            # Add an object based on these attributes
            obj = sg.Object(attr_new_object, when=self.when)
            obj = objset.add(obj, epoch_now, add_if_exist=False)

        # Return the expected inputs
        return [objset] + attr_expected_in

    def self_json(self):
        return {'when': self.when}

    def get_children_targets(self, objset):
        has_op = False
        for c in self.child:
            if isinstance(c, Operator):
                has_op = True
        if has_op:
            return super().get_children_targets(objset)
        return []

class Get(SNOperator):
    """Get attribute of an object."""

    def __init__(self, attr_type, objs):
        """Get attribute of an object.

        Args:
          attr_type: string, color, shape, or location. The type of attribute to get.
          objs: Operator instance or Object instance
        """
        super(Get, self).__init__()
        self.attr_type = attr_type
        self.objs = objs
        assert isinstance(objs, Operator)
        self.set_child(objs)

    def __str__(self):
        words = [self.attr_type, 'of', str(self.objs)]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        """Get the attribute.

        By default, get the attribute of the unique object. If there are
        multiple objects, then return INVALID.

        Args:
          objset: objset
          epoch_now: epoch now

        Returns:
          attr: Attribute instance or INVALID
        """

        if isinstance(self.objs, Operator):
            objs = self.objs(objset, epoch_now)
        else:
            objs = self.objs

        if objs is env_reg.DATA.INVALID:
            return env_reg.DATA.INVALID
        elif len(objs) != 1:
            # Ambiguous or non-existent
            return env_reg.DATA.INVALID
        else:
            attr = getattr(objs[0], self.attr_type)
            return attr

    def copy(self):
        raise NotImplementedError()

    def get_expected_input(self, should_be):
        if should_be is None:
            should_be = sg.random_attr(self.attr_type)
        objs = sg.Object([should_be])
        return [objs]


class GetObject(Get):
    def __init__(self, objs):
        super(GetObject, self).__init__('object', objs)

    def __str__(self):
        words = ['identity', 'of', str(self.objs)]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def copy(self):
        new_objs = self.objs.copy()
        return GetObject(new_objs)


class GetCategory(Get):
    def __init__(self, objs):
        super(GetCategory, self).__init__('category', objs)

    def copy(self):
        new_objs = self.objs.copy()
        return GetCategory(new_objs)


class GetViewAngle(Get):
    def __init__(self, objs):
        super(GetViewAngle, self).__init__('view_angle', objs)

    def copy(self):
        new_objs = self.objs.copy()
        return GetViewAngle(new_objs)


class GetLoc(Get):
    """Get location of object."""

    def __init__(self, objs):
        super(GetLoc, self).__init__('location', objs)

    def copy(self):
        new_objs = self.objs.copy()
        return GetLoc(new_objs)


class Exist(SNOperator):
    """Check if object with property exists."""

    def __init__(self, objs):
        super(Exist, self).__init__()
        self.objs = objs
        assert isinstance(objs, Operator)
        self.set_child(objs)

    def __str__(self):
        words = [str(self.objs), 'exist']
        if not self.parent:
            words += ['?']
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        subset = self.objs(objset, epoch_now)
        if subset == env_reg.DATA.INVALID:
            return env_reg.DATA.INVALID
        elif subset:
            # If subset is not empty
            return True
        else:
            return False

    def copy(self):
        new_objs = self.objs.copy()
        return Exist(new_objs)

    def get_expected_input(self, should_be):

        if should_be is None:
            should_be = random.random() > 0.5

        if should_be:
            should_be = [sg.Object()]
        else:
            should_be = []

        return should_be


class Switch(SNOperator):
    """Switch behaviors based on trueness of statement.

    Args:
      statement: boolean type operator
      do_if_true: operator performed if the evaluated statement is True
      do_if_false: operator performed if the evaluated statement is False
      invalid_as_false: When True, invalid statement is treated as False
      both_options_avail: When True, both do_if_true and do_if_false will be
        called the get_expected_input function
    """

    def __init__(self,
                 statement,
                 do_if_true,
                 do_if_false,
                 invalid_as_false=False,
                 both_options_avail=True):
        super(Switch, self).__init__()

        self.statement = statement
        self.do_if_true = do_if_true
        self.do_if_false = do_if_false

        self.set_child([statement, do_if_true, do_if_false])

        self.invalid_as_false = invalid_as_false
        self.both_options_avail = both_options_avail

    def __str__(self):

        words = [
            'if',
            str(self.statement) + ',',
            'then',
            str(self.do_if_true) + '?',
            'else',
            str(self.do_if_false)
        ]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        statement_true = self.statement(objset, epoch_now)
        if statement_true is env_reg.DATA.INVALID:
            if self.invalid_as_false:
                statement_true = False
            else:
                return env_reg.DATA.INVALID
        if statement_true:
            return self.do_if_true(objset, epoch_now)
        else:
            return self.do_if_false(objset, epoch_now)

    # def __getattr__(self, key):
    #     """Get attributes.

    #     The Switch operator should also have the shared attributes of do_if_true and
    #     do_if_false. Because regardless of the statement truthness, the common
    #     attributes will be true.

    #     Args:
    #       key: attribute

    #     Returns:
    #       attr1: common attribute of do_if_true and do_if_false

    #     Raises:
    #       ValueError: if the attribute is not common. Temporary.
    #     """
    #     attr1 = getattr(self.do_if_true, key)
    #     attr2 = getattr(self.do_if_false, key)
    #     if attr1 == attr2:
    #         return attr1
    #     else:
    #         raise ValueError()

    def copy(self):
        new_statement = self.statement.copy()
        new_do_if_true = self.do_if_true.copy()
        new_do_if_false = self.do_if_false.copy()
        return Switch(new_statement, new_do_if_true, new_do_if_false,
                      self.invalid_as_false, self.both_options_avail)

    def get_expected_input(self, should_be):
        """Guess objset for Switch operator.

        Here the objset is guessed based on whether or not the statement should be
        true.
        Both do_if_true and do_if_false should be executable regardless of the
        statement truthfulness. In this way, it doesn't provide additional
        information.

        Args:
          should_be: the expected output

        Returns:
          objset: updated objset
        """
        if should_be is None:
            should_be = random.random() > 0.5

        return should_be, None, None


class IsSame(SNOperator):
    """Check if two attributes are the same."""

    def __init__(self, attr1, attr2):
        """Compare to attributes.

        Args:
          attr1: Instance of Attribute or Get
          attr2: Instance of Attribute or Get

        Raises:
          ValueError: when both attr1 and attr2 are instances of Attribute
        """
        super(IsSame, self).__init__()

        self.attr1 = attr1
        self.attr2 = attr2

        self.attr1_is_op = isinstance(self.attr1, Operator)
        self.attr2_is_op = isinstance(self.attr2, Operator)

        self.set_child([self.attr1, self.attr2])

        if (not self.attr1_is_op) and (not self.attr2_is_op):
            raise ValueError('attr1 and attr2 cannot both be Attribute instances.')

        self.attr_type = self.attr1.attr_type
        assert self.attr_type == self.attr2.attr_type

    def __str__(self):
        words = [self.attr1.__str__(), 'equals', self.attr2.__str__()]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        attr1 = self.attr1(objset, epoch_now)
        attr2 = self.attr2(objset, epoch_now)

        if (attr1 is env_reg.DATA.INVALID) or (attr2 is env_reg.DATA.INVALID):
            return env_reg.DATA.INVALID
        else:
            return attr1 == attr2

    def __eq__(self, other):
        if not isinstance(other, IsSame):
            return False
        for c, o_c in zip(
                sorted(self.child, key=lambda o: o.__class__.__name__),
                sorted(other.child, key=lambda o: o.__class__.__name__)
        ):
            if c != o_c:
                return False
        return True

    def __hash__(self):
        if self.child:
            c_s = sorted(self.child, key=lambda o: o.__class__.__name__)
            return hash(tuple(
                [self.__class__.__name__] +
                [c for c in c_s]
            ))
        else:
            return hash(self.__class__.__name__)

    def copy(self):
        new_attr1 = self.attr1.copy()
        new_attr2 = self.attr2.copy()
        return IsSame(new_attr1, new_attr2)

    def get_expected_input(self, should_be, objset, epoch_now):
        if should_be is None:
            should_be = random.random() > 0.5
        # Determine which attribute should be fixed and which shouldn't

        attr1_value = self.attr1(objset, epoch_now)
        attr2_value = self.attr2(objset, epoch_now)

        attr1_fixed = attr1_value is not env_reg.DATA.INVALID
        attr2_fixed = attr2_value is not env_reg.DATA.INVALID

        if attr1_fixed:
            assert attr1_value.has_value()

        if attr1_fixed and attr2_fixed:
            # do nothing
            attr1_assign, attr2_assign = Skip(), Skip()
        elif attr1_fixed and not attr2_fixed:
            attr1_assign = Skip()
            attr2_assign = attr1_value if should_be else sg.another_attr(attr1_value)
        elif not attr1_fixed and attr2_fixed:
            attr1_assign = attr2_value if should_be else sg.another_attr(attr2_value)
            attr2_assign = Skip()
        else:
            if self.attr_type == 'view_angle':
                # if sample another view_angle if not should be
                # check for how many view_angles exist for the object first
                obj = sg.random_attr('object')
                while len(self.stim_data.ALLVIEWANGLES[obj.category.value][obj.value]) < 2:
                    obj = sg.random_attr('object')
                attr = sg.random_attr('view_angle')
                attr1_assign = attr
                attr2_assign = attr if should_be else sg.another_attr(attr)
            elif self.attr_type == 'location':
                attr = sg.random_attr('location')
                attr1_assign = attr
                attr2_assign = attr.space.sample_loc() if should_be else sg.another_attr(attr)
            else:
                attr = sg.random_attr(self.attr_type)
                attr1_assign = attr
                attr2_assign = attr if should_be else sg.another_attr(attr)
        return attr1_assign, attr2_assign


class NotSame(SNOperator):
    """Check if two attributes are not the same."""

    def __init__(self, attr1, attr2):
        """Compare to attributes.

        Args:
          attr1: Instance of Attribute or Get
          attr2: Instance of Attribute or Get

        Raises:
          ValueError: when both attr1 and attr2 are instances of Attribute
        """
        super(NotSame, self).__init__()

        self.attr1 = attr1
        self.attr2 = attr2

        self.attr1_is_op = isinstance(self.attr1, Operator)
        self.attr2_is_op = isinstance(self.attr2, Operator)

        self.set_child([self.attr1, self.attr2])

        if (not self.attr1_is_op) and (not self.attr2_is_op):
            raise ValueError('attr1 and attr2 cannot both be Attribute instances.')

        self.attr_type = self.attr1.attr_type
        assert self.attr_type == self.attr2.attr_type

    def __str__(self):
        words = [self.attr1.__str__(), 'not equals', self.attr2.__str__()]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        attr1 = self.attr1(objset, epoch_now)
        attr2 = self.attr2(objset, epoch_now)

        if (attr1 is env_reg.DATA.INVALID) or (attr2 is env_reg.DATA.INVALID):
            return env_reg.DATA.INVALID
        else:
            return attr1 != attr2

    def __eq__(self, other):
        if not isinstance(other, NotSame):
            return False
        for c, o_c in zip(
                sorted(self.child, key=lambda o: o.__class__.__name__),
                sorted(other.child, key=lambda o: o.__class__.__name__)
        ):
            if c != o_c:
                return False
        return True

    def __hash__(self):
        if self.child:
            c_s = sorted(self.child, key=lambda o: o.__class__.__name__)
            return hash(tuple(
                [self.__class__.__name__] +
                [c for c in c_s]
            ))
        else:
            return hash(self.__class__.__name__)

    def copy(self):
        new_attr1 = self.attr1.copy()
        new_attr2 = self.attr2.copy()
        return NotSame(new_attr1, new_attr2)

    def get_expected_input(self, should_be, objset, epoch_now):
        if should_be is None:
            should_be = random.random() > 0.5
        # Determine which attribute should be fixed and which shouldn't
        attr1_value = self.attr1(objset, epoch_now)
        attr2_value = self.attr2(objset, epoch_now)

        attr1_fixed = attr1_value is not env_reg.DATA.INVALID
        attr2_fixed = attr2_value is not env_reg.DATA.INVALID

        if attr1_fixed:
            assert attr1_value.has_value()

        if attr1_fixed and attr2_fixed:
            # do nothing
            attr1_assign, attr2_assign = Skip(), Skip()
        elif attr1_fixed and not attr2_fixed:
            attr1_assign = Skip()
            attr2_assign = sg.another_attr(attr1_value) if should_be else attr1_value
        elif not attr1_fixed and attr2_fixed:
            attr1_assign = sg.another_attr(attr2_value) if should_be else attr2_value
            attr2_assign = Skip()
        else:
            if self.attr_type == 'view_angle':
                # if sample another view_angle if should be
                # check for how many view_angles exist for the object first

                attr = sg.random_attr('view_angle')
                while len(self.stim_data.ALLVIEWANGLES[attr.category.value][attr.value]) < 2:
                    attr = sg.random_attr('view_angle')
                attr1_assign = attr
                attr2_assign = sg.another_attr(attr) if should_be else attr
            elif self.attr_type == 'location':
                attr = sg.random_attr('location')
                attr1_assign = attr
                attr2_assign = sg.another_attr(attr) if should_be else attr.space.sample_loc()
            else:
                attr = sg.random_attr(self.attr_type)
                attr1_assign = attr
                attr2_assign = sg.another_attr(attr) if should_be else attr
        return attr1_assign, attr2_assign


class And(SNOperator):
    """And operator."""

    def __init__(self, op1, op2):
        super(And, self).__init__()
        self.op1, self.op2 = op1, op2
        self.set_child([op1, op2])

    def __str__(self):
        words = [str(self.op1), 'and', str(self.op2)]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        return self.op1(objset, epoch_now) and self.op2(objset, epoch_now)

    def copy(self):
        new_op1 = self.op1.copy()
        new_op2 = self.op2.copy()
        return And(new_op1, new_op2)

    def get_expected_input(self, should_be, objset, epoch_now):
        # should_be is the expected output
        if should_be is None:
            should_be = random.random() > 0.5

        if should_be:
            op1_assign, op2_assign = True, True
        else:
            r = random.random()
            # Assume the two operators are independent with P[op=True]=sqrt(0.5)
            # Here show the conditional probability given the output is False
            if r < 0.414:  # 2*sqrt(0.5) - 1
                op1_assign, op2_assign = False, True
            elif r < 0.828:
                op1_assign, op2_assign = True, False
            else:
                op1_assign, op2_assign = False, False

        return op1_assign, op2_assign


class Or(SNOperator):
    """Or operator."""

    def __init__(self, op1, op2):
        super(Or, self).__init__()
        self.op1, self.op2 = op1, op2
        self.set_child([op1, op2])

    def __str__(self):
        words = [str(self.op1), 'or', str(self.op2)]
        if not self.parent:
            words[-1] += '?'
        return ' '.join(words)

    def __call__(self, objset, epoch_now):
        return self.op1(objset, epoch_now) or self.op2(objset, epoch_now)

    def copy(self):
        new_op1 = self.op1.copy()
        new_op2 = self.op2.copy()
        return Or(new_op1, new_op2)

    def get_expected_input(self, should_be, objset, epoch_now):
        # should_be is the expected output
        # should_be is True or False
        if should_be is None:
            should_be = random.random() > 0.5

        if should_be:
            r = random.random()
            # Assume the two operators are independent with P[op=False]=1-sqrt(0.5)
            # Here show the conditional probability given the output is True
            if r < 0.414:
                op1_assign, op2_assign = False, True
            elif r < 0.828:
                op1_assign, op2_assign = True, False
            else:
                op1_assign, op2_assign = True, True
        else:
            op1_assign, op2_assign = False, False
        return op1_assign, op2_assign


class SNTask(Task):

    def _add_all_nodes(self, op: Union[Operator, sg.SNAttribute], visited: dict, G: nx.DiGraph, count: int):
        visited[op] = True
        parent = count
        node_label = type(op).__name__
        G.add_node(parent, label=node_label)
        if node_label == 'Switch':
            conditional_op, if_op, else_op = op.child[0], op.child[1], op.child[2]
            conditional_root = count + 1
            _, if_root = self._add_all_nodes(conditional_op, visited, G, conditional_root)
            _, else_root = self._add_all_nodes(if_op, visited, G, if_root + 1)
            _, else_node = self._add_all_nodes(else_op, visited, G, else_root + 1)
            G.add_edge(else_root, parent)
            G.add_edge(else_node, parent)
            G.add_edge(parent, conditional_root)
            return G, count
        else:
            for c in op.child:
                if isinstance(c, Operator) and not visited[c]:
                    child = count + 1
                    _, count = self._add_all_nodes(c, visited, G, child)
                    G.add_edge(parent, child)
                else:
                    if node_label != 'Select':
                        child = count + 1
                        G.add_node(child, label='CONST')
                        G.add_edge(parent, child)
                        count += 1
        return G, count

    def guess_objset(
            self,
            objset: sg.ObjectSet,
            epoch_now: int,
            should_be: Dict = None,
            temporal_switch: bool = False
    ):
        """
        main function for generating frames based on task graph structure
        iterate through each node in topological order, and propagate the expected inputs from
        predecessors to the successors
        :return: the updated objset after going through the task graph
        """
        nodes = self.topological_sort()
        should_be_dict = defaultdict(lambda: None)

        # if should_be is None, then the output is randomly sampled
        if should_be is not None:
            should_be_dict[nodes[0]] = should_be

        # iterate over all nodes in topological order
        # while updating the expected input from the successors/children of the current node
        for node in nodes:
            should_be = should_be_dict[node]
            # checking the type of operator
            if isinstance(should_be, Skip):
                inputs = Skip() if len(node.child) == 1 else [Skip()] * len(node.child)
            elif isinstance(node, Select):
                inputs = node.get_expected_input(should_be, objset, epoch_now)
                objset = inputs[0]
                inputs = inputs[1:]
            elif (isinstance(node, IsSame) or isinstance(node, NotSame)
                  or isinstance(node, And) or isinstance(node, Or)):
                inputs = node.get_expected_input(should_be, objset, epoch_now)
            else:
                inputs = node.get_expected_input(should_be)
            # e.g. node = IsSame, should_be = True,
            # expected_input is the output of the children operators

            # makes sure outputs is a list, if node is select, then get_expected_input adds object to objset
            if len(node.child) == 1:
                outputs = [inputs]
            else:
                outputs = inputs

            if isinstance(node, Switch) and temporal_switch:
                children = node.child
                if random.random() > 0.5:
                    children.pop(1)
                else:
                    children.pop(2)
            else:
                children = node.child

            # Update the should_be dictionary for the children
            for c, output in zip(children, outputs):
                if not isinstance(c, Operator):  # if c is not an Operator
                    continue
                if isinstance(output, Skip):
                    should_be_dict[c] = Skip()
                    continue
                if should_be_dict[c] is None:
                    # If not assigned, assign
                    should_be_dict[c] = output
                # if child is an operator and there's already assigned expected output
                else:
                    # If assigned, for each object, try to merge them
                    # currently, only selects should have pre-assigned output
                    if isinstance(c, Select):
                        # Loop over new output
                        for o in output:
                            assert isinstance(o, sg.Object)
                            merged = False
                            # Loop over previously assigned outputs
                            for s in should_be_dict[c]:
                                # Try to merge
                                merged = s.merge(o)
                                if merged:
                                    break
                            if not merged:
                                should_be_dict[c].append(o)
                    else:
                        raise NotImplementedError(f'class {type(c)} not implemented')
        return objset


class TemporalTask(SNTask):
    def __init__(self, operator=None, n_frames=None, first_shareable=None, whens=None):
        super(TemporalTask, self).__init__(operator)
        self.n_frames = n_frames

        self._first_shareable = first_shareable
        self.whens = whens

    def copy(self):
        # duplicate the task
        new_task = TemporalTask()
        new_task.n_frames = self.n_frames
        new_task._first_shareable = self.first_shareable
        new_task._operator = self._operator.copy()
        return new_task

    @property
    def first_shareable(self):
        """

        :return: the frame at which the task is first shareable.
        if the task is non-shareable, first_shareable = len(task)
        if no input, start at random frame, including the possibility of non-shareable
        """
        if self._first_shareable is None:
            self._first_shareable = int(np.random.choice(np.arange(0, self.n_frames + 1)))
        return self._first_shareable

    @property
    def instance_size(self):
        # depending on the number of stimuli to sample from,
        # instance size determines the number of variations of the task
        raise NotImplementedError

    @staticmethod
    def filter_selects(self, lastk=None) -> List[Select]:
        # filter for select operators that corresponds directly to an object and match lastk
        selects = list()

        for node in self.topological_sort():
            if isinstance(node, Select) and node.check_attrs():
                # print("what is node.when:", node.when)
                if lastk and node.when == lastk:
                    selects.append(node)
                else:
                    continue  # XLEI: only keep select operators that matches certain timestamp[lastk]
                    # selects.append(node)
        return selects

    def get_relevant_attribute(self, lastk: str):
        # return the attribute of the object that is not randomly selected on lastk frame
        # for merging check purpose
        attrs = set()

        for lastk_select in self.filter_selects(self, lastk):
            for parent_op in lastk_select.parent:
                if isinstance(parent_op, Get):
                    attrs.add(parent_op.attr_type)
                elif isinstance(parent_op, Exist):
                    for attr in env_reg.DATA.ATTRS:
                        if getattr(lastk_select, attr).value:
                            attrs.add(attr)
                elif isinstance(parent_op, IsSame):
                    for op in parent_op.child:
                        if not isinstance(op, Operator):
                            attrs.add(op.attr_type)

        return attrs

    def reinit(self, copy, objs: List[sg.Object], lastk: str) -> List[sg.Object]:
        """update the task's selects in-place based on provided objects.

        :param lastk: the frame index with respect to the ending frame
        .g. if len(frames) is 3, last frame is last0, first frame is last2
        :param objs: list of objects that the selects need to match
        :type copy: copy of the TemporalTask
        :return: None if there are no leaf selects, list of objs otherwise
        """

        assert all([o.when == objs[0].when for o in objs])

        # find selects that match the lastk
        copy_filter_selects = copy.filter_selects(copy, lastk)
        filter_selects = self.filter_selects(self, lastk)

        # uncomment if multiple stim per frame
        # assert len(filter_selects) == len(copy_filter_selects)

        filter_objs = list()
        if filter_selects:
            if len(objs) < len(filter_selects):
                print('Not enough objects for select')
                return list()
            # match objs on that frame with the number of selects
            filter_objs = random.sample(objs, k=len(filter_selects))
            # print("what is filter_objs:", filter_objs)
            for select, select_copy, obj in zip(filter_selects, copy_filter_selects, filter_objs):
                # each select that matches the lastk needs to change its attrs
                # to point to the reused object from existing frames
                select.soft_update(obj)
                select_copy.hard_update(obj)
        # print("after updating, what is filter_objs:", filter_objs)
        return filter_objs

    def generate_objset(self, *args, **kwargs):
        """Guess objset for all n_epoch. this is 1 trial

        Returns:
          objset: full objset for all n_epoch
        """

        n_epoch = self.n_frames
        objset = sg.ObjectSet(n_epoch=n_epoch)

        # Guess objects
        # Importantly, we generate objset backward in time
        ## xlei: only update the last epoch
        epoch_now = n_epoch - 1

        objset = self.guess_objset(objset, epoch_now, *args, **kwargs)
        return objset

    def to_json(self):
        info = dict()
        info['n_frames'] = int(self.n_frames)
        info['first_shareable'] = int(self.first_shareable)
        info['whens'] = self.whens
        info['operator'] = self._operator.to_json()
        return info

    def to_graph(self):
        G = nx.DiGraph()
        visited = defaultdict(lambda: False)
        G, _ = self._add_all_nodes(self._operator, visited, G, 0)
        return G

    def draw_graph(self, fp, G: nx.DiGraph = None):
        if G is None:
            G = self.to_graph()
        G = G.reverse()
        A = nx.nx_agraph.to_agraph(G)
        A.draw(fp, prog='dot')
        return


TASK = Tuple[Union[Operator, sg.SNAttribute], TemporalTask]
GRAPH_TUPLE = Tuple[nx.DiGraph, int, int]


def all_subclasses(cls):
    """
    function to retrieve all subclasses of a class
    @param cls: class
    @return:
    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def get_operator_dict() -> Dict[str, Callable]:
    """
    retrieve operator class names and their classes
    @return: dictionary of format {str: SNOperator class}
    """
    # function to retrieve class names and their classes
    return {cls.__name__: cls for cls in all_subclasses(SNOperator)}


def graph_to_operators(
        G: nx.DiGraph,
        root: int,
        operators: Dict[int, str],
        operator_families: Dict[str, Callable],
        whens: List[str],
) -> Union[Operator, sg.SNAttribute]:
    """
    given a task graph G, convert it into an operator that is already nested with its successors
    this is done by traversing the graph
    :param G: the task graph, nx.Graph
    :param root: the node number of the root operator
    :param operators: dictionary for converting node number to operator name
    :param operator_families: dict of class names and their classes
    :param whens: each select is associated with 1 when
    :return:
    """
    children = list(G.successors(root))
    if children:
        # check the root's type of operator
        if operators[root] == 'Select':
            attr_dict = {attr: None for attr in env_reg.DATA.ATTRS}
            for child in children:
                # each child is an int indicating the node number
                if 'Get' in operators[child]:
                    attr = operators[child].split('Get')[1].lower()
                    attr = 'view_angle' if attr == 'viewangle' else attr
                    attr_dict[attr] = graph_to_operators(G, child, operators, operator_families, whens)
                else:
                    raise ValueError(f'Select cannot have {operators[child]} as a child operator')
            return Select(**attr_dict, when=whens[root])
        elif operators[root] == 'Exist':
            return Exist(graph_to_operators(G, children[0], operators, operator_families, whens))
        elif 'Get' in operators[root]:
            # init a Get operator
            return operator_families[operators[root]](
                graph_to_operators(G, children[0], operators, operator_families, whens))
        elif operators[root] in env_reg.DATA.LOGIC_OPS:
            # init a boolean operator
            assert len(children) > 1
            ops = [graph_to_operators(G, c, operators, operator_families, whens) for c in children]
            if isinstance(ops[0], sg.SNAttribute) or isinstance(ops[1], sg.SNAttribute):
                if isinstance(ops[0], sg.SNAttribute):
                    attr = ops[0]
                    op = ops[1]
                    attr_i = 0
                else:
                    attr = ops[1]
                    op = ops[0]
                    attr_i = 1
                if attr.attr_type != op.attr_type:
                    ops[attr_i] = sg.random_attr(op.attr_type)
            return operator_families[operators[root]](ops[0], ops[1])
        else:
            raise ValueError(f"Unknown Operator {operators[root]}")
    else:
        if operators[root] == 'Select':
            return Select(when=whens[root])
        else:
            return sg.random_attr(random.choice(env_reg.DATA.ATTRS))


def read_task(task_fp: str):
    """
    Read a task from a json file

    @param task_fp: the file path to the task
    @return: a TemporalTask instance
    """
    with open(task_fp, 'r') as f:
        task_info = json.load(f)

    # first load the operator objects
    task_info['operator'] = load_operator_json(task_info['operator'])

    # reinitialize using the parent task class. (the created task object is functionally identical)
    task = TemporalTask(
        operator=task_info['operator'],
        n_frames=task_info['n_frames'],
        first_shareable=task_info['first_shareable'],
        whens=task_info['whens']
    )
    return task


def load_operator_json(
        op_dict: dict,
        operator_families: Dict[str, Callable] = None,
        attr_families: Dict[str, Callable] = None,
) -> Union[Operator, sg.SNAttribute]:
    """
    given json dictionary, convert it into an operator that is already nested with its successors
    :param op_dict: the operator dictionary
    :param operator_families: dict of operator class names and their class init functions
    :param attr_families: dict of attribute class names and their class init functions
    :return: initialized Operator
    """
    if operator_families is None:
        operator_families = get_operator_dict()
    if attr_families is None:
        attr_families = sg.get_attr_dict()
    name = op_dict['name']
    if not 'value' in op_dict:
        children: List[dict] = op_dict['child']
        # check the type of operator
        if name == 'Select':
            attr_dict = {attr: None for attr in env_reg.DATA.ATTRS}
            for d in children:
                if not 'value' in d:
                    if 'Get' in d['name']:
                        attr = d['name'].split('Get')[1].lower()
                        attr = 'view_angle' if attr == 'viewangle' else attr
                        attr_dict[attr] = load_operator_json(children[0], operator_families, attr_families)
                    else:
                        raise ValueError(f'Select cannot have {children[0]["name"]} as a child operator')
            return Select(**attr_dict, when=op_dict['when'])
        elif name == 'Exist':
            return Exist(load_operator_json(children[0], operator_families, attr_families))
        elif 'Get' in name:
            # init a Get operator
            return operator_families[name](
                load_operator_json(children[0], operator_families, attr_families)
            )
        elif name in env_reg.DATA.LOGIC_OPS:
            # init a boolean operator
            assert len(children) > 1
            ops = [load_operator_json(c, operator_families, attr_families) for c in children]
            if isinstance(ops[0], sg.SNAttribute) or isinstance(ops[1], sg.SNAttribute):
                if isinstance(ops[0], sg.SNAttribute):
                    attr = ops[0]
                    op = ops[1]
                    attr_i = 0
                else:
                    attr = ops[1]
                    op = ops[0]
                    attr_i = 1
                # match the attribute type of the operator
                if attr.attr_type != op.attr_type:
                    ops[attr_i] = sg.random_attr(op.attr_type)
            return operator_families[name](ops[0], ops[1])
        elif name == 'Switch':
            assert len(children) == 3
            statement, do_if_true, do_if_false = [
                load_operator_json(d, operator_families, attr_families)
                for d in children]
            return Switch(statement, do_if_true, do_if_false)
        else:
            raise ValueError(f"Unknown Operator {name}")
    else:
        # we have reached an attribute
        init = {}
        if 'category' in op_dict:
            init['category'] = load_operator_json(op_dict['category'], operator_families, attr_families)
        elif 'sn_object' in op_dict:
            init['sn_object'] = load_operator_json(op_dict['sn_object'], operator_families, attr_families)

        return attr_families[name](value=op_dict['value'], **init)


def subtask_generation(
        env_spec: SNEnvSpec,
        subtask_graph: GRAPH_TUPLE,
        node_label_dict: dict = None,
        existing_whens: dict = None
) -> Tuple[TASK, dict]:
    """
    generate a TemporalTask from a subtask graph
    @param env_spec: data class used to specify number of delay frames
    @param subtask_graph: the task graph
    @param node_label_dict: dictionary of node number to node label e.g. 1: 'Select'
    @param existing_whens: for combining multiple tasks together, a dictionary of the format {select: 'lastk'}
    @return:
    """
    existing_whens = dict() if not existing_whens else existing_whens
    # avoid duplicate whens across subtasks during switch generation
    subtask_G, root, _ = subtask_graph
    operator_families = get_operator_dict()

    if node_label_dict is None:
        node_label_dict = {node[0]: node[1]['label'] for node in subtask_G.nodes(data=True)}
    selects = [op for op in subtask_G.nodes() if 'Select' == node_label_dict[op]]

    whens = env_spec.check_whens(env_spec.sample_when(len(selects)), list(existing_whens.values()))
    n_frames = env_reg.compare_when(whens) + 1  # find highest lastk to determine number of frames in task

    whens = {select: when for select, when in zip(selects, whens)}
    existing_whens.update(whens)
    assert len(existing_whens.values()) == len(set(existing_whens.values())), 'whens are duplicated'
    op = graph_to_operators(subtask_G, root, node_label_dict, operator_families, whens)
    return (op, TemporalTask(operator=op, n_frames=n_frames, whens=whens)), existing_whens


def switch_generation(conditional: TASK, do_if: TASK, do_else: TASK, existing_whens: dict, **kwargs) -> TASK:
    """
    combines all 3 temporal tasks and initialize the switch operator
    """
    conditional_op, conditional_task = conditional
    if_op, if_task = do_if
    else_op, else_task = do_else

    op = Switch(conditional_op, if_op, else_op, **kwargs)
    n_frames = env_reg.compare_when(existing_whens.values()) + 1

    # env_reg.DATA.MAX_MEMORY = n_frames
    return op, TemporalTask(operator=op, n_frames=n_frames, whens=existing_whens)


get_family_dict = OrderedDict([
    ('category', GetCategory),
    ('object', GetObject),
    ('view_angle', GetViewAngle),
    ('location', GetLoc)
])



"""
Classes for building temporal composite tasks
"""
from __future__ import annotations

import random
import re
from collections import defaultdict
from functools import partial
from typing import Tuple, Dict, List, Set

import numpy as np

import iwisdm.envs.shapenet.registration as env_reg
from iwisdm.envs.shapenet.registration import SNStimData
import iwisdm.envs.shapenet.stim_generator as sg
import iwisdm.envs.shapenet.task_generator as tg
from iwisdm.utils.read_write import add_cross, render_stimset


class FrameInfo(object):
    def __init__(self, task: tg.TemporalTask, objset: sg.ObjectSet):
        """
        contains information about individual task's frames
        used for combining multiple temporal tasks, initialize with 1 task,
        stores each frame object in frame_list
        :param task: task_generator.TemporalTask object
        :param objset: objset related to the task
        """
        assert isinstance(objset, sg.ObjectSet)
        assert isinstance(task, tg.TemporalTask)
        if task.n_frames != objset.n_epoch:
            raise ValueError('Task epoch does not equal objset epoch')

        n_epochs = task.n_frames
        self.objset = objset
        self.frame_list = list()  # List of Frame objects
        self.n_epochs = n_epochs
        self.first_shareable = task.first_shareable
        # TODO: decide if task_question needs to be kept
        # task_question = str(task)
        # task_answers = [env_reg.get_target_value(t) for t in task.get_target(objset)]

        for i in range(n_epochs):
            # keep track of which frames are the ending and starting of which tasks
            description = list()
            if i == self.n_epochs - 1:
                description.append("ending of task %d" % 0)
            else:
                if i == 0:
                    description.append("start of task %d" % 0)
            self.frame_list.append(
                self.Frame(
                    fi=self,
                    idx=i,
                    relative_tasks={0},
                    description=description
                )
            )
        if objset:
            # iterate all objects in objset and add to each frame
            for obj in objset:
                # iterate over all frames that the objset belongs to
                for epoch in range(obj.epoch[0], obj.epoch[1]):
                    self.frame_list[epoch].objs.append(obj.copy())

        self.last_task = 0
        self.last_task_start = 0
        self.last_task_end = len(self.frame_list) - 1

    def dump(self) -> Dict[int, Dict]:
        # get information about the frame_info
        info = {i: dict() for i, frame in enumerate(self.frame_list)}
        for i, frame in enumerate(self.frame_list):
            info[i]['description'] = frame.description
            info[i]['relative_tasks'] = list(frame.relative_tasks)
            info[i]['relative_task_epoch_idx'] = frame.relative_task_epoch_idx
        return info

    def add_new_frames(self, i, relative_tasks):
        """
        add new empty frames and update objset and p during composition
        :param i: number of new frames
        :param relative_tasks: the tasks associated with the new frames
        :return: True if epoch increased
        """
        if i <= 0:
            return False
        for j in range(i):
            self.frame_list.append(self.Frame(fi=self,
                                              idx=len(self.frame_list),
                                              relative_tasks=relative_tasks
                                              ))
        self.objset.increase_epoch(self.objset.n_epoch + i)
        return True

    def add_distractor(self, distractor: sg.Object, frame_idx: int, description: str = None):
        frame = self.frame_list[frame_idx]
        frame.objs.append(distractor)
        frame.relative_tasks.add(-1)
        if not description:
            description = 'distractor'
        frame.description.append(description)
        self.objset.add(
            obj=distractor,
            epoch_now=len(self.frame_list) - 1,
            delete_if_can=True,
            merge_idx=frame_idx
        )
        return

    @staticmethod
    def update_relative_tasks(new_frame_info, relative_tasks):
        # update the relative_task info for each frame in the new task (from 0 to new task idx) after composition
        next_task_idx = list(relative_tasks)[0]
        for frame in new_frame_info:
            frame.relative_tasks = relative_tasks
            # new_task_info should only contain 1 task
            if 0 not in frame.relative_task_epoch_idx:
                print(frame.relative_task_epoch_idx)
            frame.relative_task_epoch_idx[next_task_idx] = frame.relative_task_epoch_idx.pop(0)
            for i, description in enumerate(frame.description):
                if 'start' in description:
                    frame.description[i] = f'start of task {next_task_idx}'
                if 'ending' in description:
                    frame.description[i] = f'ending of task {next_task_idx}'
        return

    def get_start_frame(self, new_task_info: TaskInfoCompo, relative_tasks: Set[int]):
        """
        randomly sample a starting frame to start merging and add new frames if needed

        this is done by checking length of both tasks, then starting first based on first_shareable

        if both start at the same frame, but new task ends earlier,
        then force the new task to start later by adding more empty frames to the old task.
        otherwise, if new task start frame is after existing task's start frame,
        then new task can end earlier than existing task

        Overall, task order is arranged such that new task appears or finishes after the existing task
        avoid overlapping response frames
        :param new_task_info: TaskInfoCompo object of the new task
        :param relative_tasks: set of task indices used by the new task, relative to the TaskInfoCompo
        :return: index of frame to begin merging with respect to existing frame_info
        """
        assert len(new_task_info.tasks) == 1

        first_shareable = self.first_shareable
        shareable_frames = self.frame_list[first_shareable:]
        new_task_len = new_task_info.n_epochs
        new_first_shareable = new_task_info.tasks[0].first_shareable
        self.update_relative_tasks(new_task_info.frame_info, relative_tasks)

        if len(shareable_frames) == 0:
            # queue, add frames and start merging
            self.first_shareable = len(self.frame_list)
            self.add_new_frames(new_task_len, relative_tasks)

            self.last_task_end = len(self.frame_list) - 1
            self.last_task = list(relative_tasks)[0]
            self.first_shareable = new_first_shareable + first_shareable
            return first_shareable
        else:
            # add more frames
            if len(shareable_frames) == new_task_len:
                self.add_new_frames(1, relative_tasks)
                first_shareable += 1
            else:
                self.add_new_frames(new_task_len - len(shareable_frames), relative_tasks)

            shareable_frames = self.frame_list[first_shareable:]
            # first check if start of alignment
            # if the alignment starts with start of existing task, then check if new task ends earlier
            aligned = False
            while not aligned:
                # inserting t3 into T = {t1,t2} (all shareable) in cases where t1 ends before t2 but appears before t2,
                # last_task_start is after t1, last_task_end is after t1.
                # suppose shareable_frames[new_task_len] is the index of end of t1
                # then shift to the right, at next loop, t2 ends, then add new frame

                # check where the response frame of the new task would be in shareable_frames
                # if the response frames are overlapping, then shift alignment to "right"
                if first_shareable + new_task_len - 1 == self.last_task_end:
                    first_shareable += 1
                # if the sample frames are overlapping, then check if the new task ends before the last task
                # if so, then shift starting frame
                elif first_shareable == self.last_task_start \
                        and first_shareable + new_task_len - 1 <= self.last_task_end:
                    first_shareable += 1
                else:
                    aligned = True

                if len(shareable_frames) < new_task_len:
                    self.add_new_frames(1, relative_tasks)
                shareable_frames = self.frame_list[first_shareable:]

            self.last_task_end = first_shareable + new_task_len - 1
            self.last_task = list(relative_tasks)[0]
            self.last_task_start = first_shareable

            self.first_shareable = new_first_shareable + first_shareable
            return first_shareable

    def __len__(self):
        return len(self.frame_list)

    def __iter__(self):
        return self.frame_list.__iter__()

    def __getitem__(self, item):
        return self.frame_list.__getitem__(item)

    class Frame(object):
        # frame object within frame_info list, used to merge frames from different tasks
        def __init__(
                self,
                fi,
                idx: int,
                relative_tasks: Set[int],
                description: List[str] = None,
                objs: List[sg.Object] = None
        ):
            """
            :param fi: the frame info object this frame belongs to
            :param idx: the index of the frame
            :param relative_tasks: which tasks this frame belongs to
            :param description: information about the frame
            :param objs: the list of objects in this frame
            """
            assert isinstance(fi, FrameInfo)

            self.fi = fi
            self.idx = idx
            self.relative_tasks = relative_tasks
            self.description = description if description else list()
            self.objs = objs if objs else list()

            self.relative_task_epoch_idx = dict()
            for task in self.relative_tasks:
                self.relative_task_epoch_idx[task] = self.idx

        def compatible_merge(self, new_frame):
            assert isinstance(new_frame, FrameInfo.Frame)

            self.relative_tasks = self.relative_tasks | new_frame.relative_tasks  # union of two sets of tasks
            self.description = self.description + new_frame.description  # append new descriptions

            for new_obj in new_frame.objs:
                self.fi.objset.add(new_obj.copy(), len(self.fi) - 1, merge_idx=self.idx)  # update objset
            self.objs = self.fi.objset.dict[self.idx].copy()

            # update the dictionary for relative_task_epoch
            temp = self.relative_task_epoch_idx.copy()
            temp.update(new_frame.relative_task_epoch_idx)
            self.relative_task_epoch_idx = temp

        def __str__(self):
            return 'frame: ' + str(self.idx) + ', relative tasks: ' + \
                ','.join([str(i) for i in self.relative_tasks]) \
                + ' objects: ' + ','.join([str(o) for o in self.objs])


class TaskInfoCompo(object):
    """
    Storage of composition information,
    including task_frame_info, task_example, task and objset
    correct usage: init objset and frame_info first, then generate TaskInfoCompo

    :param frame_info: FrameInfo
    :param task: task family
    """

    def __init__(self, task: tg.TemporalTask, frame_info: FrameInfo):
        # combining with a second task should be implemented incrementally

        self.task_objset = dict()
        self.tasks = [task]  # list of tasks in composition
        self.changed = list()
        self.frame_info = frame_info
        self.reset()

    def reset(self):
        if self.frame_info is None:
            objset = self.task.generate_objset()
            self.frame_info = FrameInfo(self.task, objset)

        self.task_objset[0] = self.frame_info.objset.copy()
        self.frame_info = self.frame_info
        self.tempo_dict = dict()

    def merge(self, new_task_info):
        """
        temporally combine the two tasks while reusing stimuli
        removed previous implementation where stimuli is not reused
        :param new_task_info: the new task to be temporally combined with the current task
        :return:
        """

        new_task = new_task_info.tasks[0]
        new_task_copy: tg.TemporalTask = new_task.copy()

        new_task_idx = len(self.tasks)
        start_frame_idx = self.frame_info.get_start_frame(new_task_info, relative_tasks={new_task_idx})

        # init new ObjSet to combine new and old
        objset = sg.ObjectSet(n_epoch=new_task_copy.n_frames)
        changed = False
        # change the necessary selects first
        for i, (old_frame, new_frame) in enumerate(zip(self.frame_info[start_frame_idx:], new_task_info.frame_info)):
            last_k = 'last%d' % (len(new_task_info.frame_info) - i - 1)
            # if there are objects in both frames, then update the new task's selects
            if old_frame.objs and new_frame.objs:
                # update the select such that it corresponds to the same object
                # checks how many selects and if there are enough objects for the selects
                filter_objs = new_task.reinit(new_task_copy, old_frame.objs, last_k)
                if filter_objs:
                    changed = True
                    for obj in filter_objs:
                        objset.add(obj=obj.copy(), epoch_now=new_task_copy.n_frames - 1, merge_idx=i)
                else:
                    raise RuntimeError('Unable to reuse')

        # get new objset for new task based on updated operators and merge
        if changed:
            # update objset based on existing objects, guess_objset resolves conflicts
            new_objset = new_task_copy.guess_objset(objset, new_task_copy.n_frames - 1)
            updated_fi = FrameInfo(new_task_copy, new_objset)  # initialize new frame info based on updated objset
            FrameInfo.update_relative_tasks(updated_fi, relative_tasks={new_task_idx})
            for i, (old_frame, new_frame) in enumerate(zip(self.frame_info[start_frame_idx:], updated_fi)):
                old_frame.compatible_merge(new_frame)  # update frame information

            self.changed.append((new_task_idx, new_task, new_task_copy))
            self.task_objset[new_task_idx] = new_objset  # update the per task objset dictionary
        else:
            for i, (old_frame, new_frame) in enumerate(
                    zip(self.frame_info[start_frame_idx:], new_task_info.frame_info)):
                old_frame.compatible_merge(new_frame)
            self.task_objset[new_task_idx] = new_task_info.task_objset[0]  # update the per task objset dictionary
        self.tasks.append(new_task)
        return

    def temporal_switch(self, task_info1, task_info2, switch_first=None):
        assert isinstance(task_info1, TaskInfoCompo)
        assert isinstance(task_info2, TaskInfoCompo)

        if any(task.is_bool_output() for task in [self.tasks[-1], task_info1.tasks[-1], task_info2.tasks[-1]]):
            raise ValueError('Switch tasks must have boolean outputs')

        self.tempo_dict['self'] = self.get_task_info_dict()[1]
        self.tempo_dict['task1'] = task_info1.get_task_info_dict()[1]
        self.tempo_dict['task2'] = task_info2.get_task_info_dict()[1]

        if not switch_first:
            answer = self.tempo_dict['self']['answers'][len(self.frame_info) - 1]
            if answer not in ['true', 'false']:
                raise ValueError('End of task is not boolean output')
            switch_first = True if answer == 'true' else False
        task_info = task_info1 if switch_first else task_info2

        self.frame_info.first_shareable = len(self.frame_info)
        self.merge(task_info)
        return

    def get_obj_info_dict(self):
        obj_info = defaultdict(list)  # key: epoch, value: list of dictionary of object info
        count = 0
        for epoch, objs in sorted(self.frame_info.objset.dict.items()):
            num_objs = len(objs)
            if num_objs > 0:
                count += 1
            for obj in objs:
                # if not obj.deletable or num_objs == 1:
                #     # remove this if instruction says observe distractor with add_attr
                #     count += 1
                info = dict()
                info['count'] = count
                info['obj'] = obj
                info['tasks'] = set()  # tasks that involve the stim
                info['attended_attr'] = defaultdict(set)  # key: task, value: attribute of the stim that are relevant
                obj_info[epoch].append(info)
        return obj_info

    def get_instruction_obj_info(self) -> Tuple[str, Dict[int, List]]:
        """ recompiles task instructions based on the composition.
        e.g. observe object1, delay, task1 instruction, delay, observe object2, task2 instruction.
        adds delay instructions during empty frames
        :return: changed task instruction and obj_info dictionary
        """
        obj_info = self.get_obj_info_dict()
        cur = 0

        # todo: uncomment below to allow instruction generation
        compo_instruction = ''
        if self.tempo_dict:
            compo_instruction += 'compo task 1: '

        # for each frame, find the tasks that use the stimuli. For each of these tasks,
        # compare the object in task_objset with obj, and rename lastk with ith object
        was_delay = False
        for epoch, frame in enumerate(self.frame_info):
            add_delay = True
            if obj_info[epoch]:  # if the frame contains stim/objects
                if len(obj_info[epoch]) == 1:
                    info_dict = obj_info[epoch][0]
                    compo_instruction += f'observe object {info_dict["count"]}, '
                else:  # if there are multiple stimuli in the frame (i.e. distractor)
                    distractor_attr = [
                        d.split('distractor with different ')[1]
                        for d in frame.description if 'distractor with different ' in d
                    ]
                    if not distractor_attr:
                        raise RuntimeError('No distractor description')

                    for info_dict in obj_info[epoch]:
                        if info_dict['obj'].deletable:
                            # remove this if instruction says observe distractor with add_attr
                            continue
                        add_attr, loc_attr = '', ''
                        for attr in distractor_attr:
                            if attr == 'location':
                                loc_attr += f' at {attr}: {getattr(info_dict["obj"], attr)}'
                            else:
                                add_attr += f' with {attr}: {getattr(info_dict["obj"], attr)}'
                        compo_instruction += f'observe object {info_dict["count"]}{add_attr}{loc_attr}, '
                add_delay, was_delay = False, False

            for d in frame.description:
                if 'ending' in d:  # align object count with lastk for each task
                    task_idx = int(re.search(r'\d+', d).group())
                    task_instruction = str(self.tasks[task_idx])

                    # find epoch relative to ending task
                    lastks = [m for m in re.finditer('last\d+', task_instruction)]
                    # TODO: modify this for multiple stim per frame
                    for lastk in lastks:
                        k = int(re.search(r'\d+', lastk.group()).group())
                        i = epoch - k
                        relative_i = frame.relative_task_epoch_idx[task_idx] - k
                        match = self.compare_objs(obj_info[i], self.task_objset[task_idx].dict[relative_i])
                        if match:
                            task_instruction = re.sub(
                                f'last{k}.*?object',
                                f'object {match["count"]}',
                                task_instruction
                            )
                            match['tasks'].add(task_idx)
                            match['attended_attr'][task_idx] = match['attended_attr'][task_idx].union(
                                self.tasks[task_idx].get_relevant_attribute(f'last{k}'))
                            cur += 1
                        else:
                            raise RuntimeError('No match')
                    compo_instruction += (task_instruction + ' ') \
                        if epoch != len(self.frame_info) - 1 else task_instruction
                    add_delay, was_delay = False, False

                    if self.tempo_dict:
                        # in temporal switch, if at the end of the original task,
                        # then add the conditional texts in the instruction
                        if epoch == len(self.tempo_dict['self']['answers']) - 1:
                            compo_instruction += ' if end of compo task 1 is true, then do compo task 2: '
                            compo_instruction += self.object_add(self.tempo_dict['task1']['instruction'], cur)
                            compo_instruction += 'otherwise, do compo task 3: '
                            compo_instruction += self.object_add(self.tempo_dict['task2']['instruction'], cur)
                            return compo_instruction, obj_info
            if add_delay and not was_delay:
                compo_instruction += 'delay, '
                was_delay = True
        return compo_instruction, obj_info

    def get_target_value(self, examples: List[Dict]) -> List[str]:
        """
        get the correct response for each frame
        :param examples: list of dictionary containing the information about individual tasks
        :return: the response for each frame in list. if there is no response, then null is saved
        """
        answers = list()
        for frame in self.frame_info:
            ending = False
            for d in frame.description:  # see frame_info init
                if 'ending' in d:  # if the frame is the end of any individual task
                    task_idx = int(re.search(r'\d+', d).group())  # get which task is ending
                    answers.append(examples[task_idx]['answers'][0])  # add the task answer
                    ending = True
            if not ending:
                answers.append('null')
        return answers

    def get_task_info_dict(self, is_instruction=True, external_instruction=None) -> Tuple[List[Dict], Dict]:
        """
        get task examples
        :return: tuple of list of dictionaries containing information about the requested tasks
        and dictionary of the compositional task
        """
        per_task_info = list()
        for i, task in enumerate(self.tasks):
            per_task_info.append({
                'family': str(task.__class__.__name__),
                # saving an epoch explicitly is needed because
                # there might be no objects in the last epoch.
                'epochs': int(task.n_frames),
                'question': str(task),
                'objects': [o.dump() for o in self.task_objset[i]],
                'answers': [str(env_reg.get_target_value(t)) for t in task.get_target(self.task_objset[i])],
                'first_shareable': int(task.first_shareable),
            })

        if is_instruction:
            comp_instruction, obj_info = self.get_instruction_obj_info()
            objs = list()
            for epoch, info_dicts in obj_info.items():
                for info_dict in sorted(info_dicts, key=lambda x: x['count']):
                    info_dict['obj'] = info_dict['obj'].dump()
                    info_dict['tasks'] = list(info_dict['tasks'])
                    info_dict['attended_attr'] = {k: list(v) for k, v in info_dict['attended_attr'].items()}
                    objs.append(info_dict)
        else:
            comp_instruction = external_instruction
            objs = [o.dump() for o in self.frame_info.objset]

        compo_info = {
            'epochs': int(len(self.frame_info)),
            'objects': objs,
            'instruction': comp_instruction,
            'answers': self.get_target_value(per_task_info)
        }
        return per_task_info, compo_info

    def add_distractor_time(self, n_distractor: int, stim_data: SNStimData):
        """
        add distractors to the empty frames of a trial
        in the simplest case, if original task is:
            observe object 1, delay, observe object 2,
            attr of object 1 same as attr of object 2?
        the new task will be:
            observe object 1, observe object 2, observe object 3,
            attr of object 1 same as attr of object 3?
        @param n_distractor: how many distractors to add
        (bounded by the number of empty frames in the trial)
        @return:
        """
        empty_frames = [i for i, frame in enumerate(self.frame_info) if not frame.objs]
        if n_distractor > len(empty_frames):
            n_distractor = len(empty_frames)
        frames_to_add = sorted(random.sample(empty_frames, n_distractor))
        for i in frames_to_add:
            self.frame_info.add_distractor(
                distractor=sg.Object(when=i, deletable=True),
                frame_idx=i
            )
        return

    def add_distractor_frame(self, n_distractor: int, stim_data: SNStimData):
        """
        add distractors to the sample frames of a trial
        in the simplest case, if original task is:
            observe object 1, observe object 2,
            attr of object 1 same as attr of object 2?
        the new task will be:
            observe (object 1 with attr x), observe object 2,
            diff_attr of object 1 same as diff_attr of object 3?
        e.g. attr is category, x is chair, y is car, diff_attr is location, so the new task instruction is:
            observe object 1 with category chair, observe object 3,
            location of object 1 same as location of object 3?

        @param n_distractor: how many distractors to add
        (bounded by how many tasks, and how many sample frames in each task)
        @return:
        """
        stim_frames = [i for i, frame in enumerate(self.frame_info) if len(frame.objs) == 1]
        relevant_frames = [i for i in stim_frames if -1 not in self.frame_info[i].relative_tasks]
        if n_distractor > len(relevant_frames):
            n_distractor = len(relevant_frames)

        frames_to_add = sorted(random.sample(relevant_frames, n_distractor))
        for i in frames_to_add:
            frame = self.frame_info[i]
            attrs = set()
            for task_idx in frame.relative_tasks:
                task = self.tasks[task_idx]
                task_epoch = frame.relative_task_epoch_idx[task_idx]
                attrs = attrs.union(task.get_relevant_attribute(f'last{task.n_frames - task_epoch - 1}'))
            attr_new_object = list()
            other_attrs = stim_data.ALL_ATTRS - attrs
            if not other_attrs:  # all attributes are used, cannot add distractor with different attribute
                continue

            for attr_type in other_attrs | {'location'}:
                existing_obj = frame.objs[0]
                attr_new_object.append(sg.another_attr(getattr(existing_obj, attr_type)))

            self.frame_info.add_distractor(
                distractor=sg.Object(attrs=attr_new_object, when=i, deletable=True),
                frame_idx=i,
                description=f'distractor with different {other_attrs.pop()}'
            )
        return

    def generate_trial(
            self,
            canvas_size: int,
            fixation_cue: bool,
            stim_data: SNStimData,
            return_objset: bool = False,
            add_distractor_frame: int = 0,
            add_distractor_time: int = 0,
    ) -> Tuple[List[np.ndarray], List[Dict], Dict]:
        # TODO: return copy of objset, not add distractor in place
        if add_distractor_frame > 0:
            self.add_distractor_frame(add_distractor_frame, stim_data)

        if add_distractor_time > 0:
            self.add_distractor_time(add_distractor_time, stim_data)

        objset = self.frame_info.objset
        per_task_info_dict, compo_info_dict = self.get_task_info_dict()

        imgs = []
        for i, (epoch, frame) in enumerate(zip(render_stimset(objset, canvas_size, stim_data), self.frame_info)):
            if fixation_cue:
                # add fixation cues to all frames except for task ending frames
                if not any('ending' in description for description in frame.description):
                    epoch = add_cross(epoch)
            imgs.append(epoch)

        if return_objset:
            print('returning objset')
            return imgs, per_task_info_dict, compo_info_dict, objset
        return imgs, per_task_info_dict, compo_info_dict

    def __len__(self):
        # return number of tasks involved
        return len(self.frame_info)

    @property
    def n_epochs(self):
        return len(self.frame_info)

    @staticmethod
    def compare_objs(info_dicts, l2):
        for info_dict in info_dicts:
            obj1 = info_dict['obj']
            if not obj1.deletable:
                for obj2 in l2:
                    if obj1.compare_attrs(obj2):
                        return info_dict
        return None

    @staticmethod
    def _add(match, x):
        # helper function for composite task instruction
        val = match.group()
        return re.sub(r'\d+', lambda m: str(int(m.group()) + x), val)

    def object_add(self, t: str, x: int):
        # finds all "object k" substrings, and substitute them with "object k+x"
        # see info_generator_test.testObjectAddOne() for example
        return re.sub(r'object \d+', partial(self._add, x=x), t)



"""
classes for registering a new environment
"""

import glob
import os
from typing import Tuple, Iterable, Dict
from collections import OrderedDict

import numpy as np
from numpy.typing import NDArray
import pandas as pd


class Constant:
    """
    singleton class, constants for the environment
    """

    BOOL_OP: Iterable = None
    LOGIC_OP: Iterable = None
    ATTRS: Iterable = None

    def __init__(self, **kwargs):
        return


class EnvSpec:
    """
        Specification for any environment, including the grid size,
        maximum number of delay frames in a task, the probability of adding delay frames.

        Also includes the specification for automatic random task generation,
        this includes maximum number of operators in the task, maximum depth of the task,
        the root operator of the task, and the generation configuration.

        see iwisdm/envs/shapenet/registration.py for example
        Args:
            grid_size: convert image canvas into (n, n) grid for location tasks
            max_delay: upper bound on the number of delay frames in a task
            delay_prob: how likely a delay frame is added, [0.0, 1.0]
            auto_gen_config: AutoTask generation configuration, see iwisdm/envs/shapenet/registration.py for example
        """

    def __init__(
            self,
            grid_size: Tuple[int, int] = (2, 2),
            max_delay: int = 5,
            delay_prob=None,
            auto_gen_config: Dict = None,
            **kwargs
    ):
        """
        environment trial generation specifications

        @param grid_size: convert image canvas into (n, n) grid for location tasks
        @param max_delay: upper bound on the number of delay frames in a task
        @param delay_prob: how likely a delay frame is added, [0.0, 1.0]
        @param auto_gen_config:
        @param kwargs:
        """
        self.MAX_DELAY = max_delay
        self.delay_prob = delay_prob
        self.auto_gen_config = auto_gen_config
        self.grid = get_grid(grid_size)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_grid_key(self, space):
        """
        convert from space to grid coordinate
        @param space:
        @return:
        """
        # convert from space to grid coordinate
        return list(self.grid.keys())[list(self.grid.values()).index(list(space.value))]


class StimData:
    """
    class for storing and retrieving the stimuli dataset information
    Args:
        dir_path: file path to the stimuli dataset. the folder has subdir structure:
            dir_path:
                -train
                    -imgs
                    -train_metadata.{csv/pkl}
                -val
                    -imgs
                    -val.{csv/pkl}
                -test
                    -imgs
                    -test_metadata.{csv/pkl}
        find_subdir: boolean, if True, find the subdirectories for train, validation, test splits
        df: pandas dataframe, the metadata for the stimuli dataset
        img_folder_path: file path to the stimuli images
        splits: dictionary containing information to the train, validation, test splits
    """

    def __init__(
            self,
            dir_path: str = None,
            find_subdir: bool = True,
            df: pd.DataFrame = None,
            img_folder_path: str = None,
            splits: Dict = None
    ):
        self.dir_path = dir_path
        self.find_subdir = find_subdir
        if not find_subdir:
            assert img_folder_path is not None and df is not None, \
                'img_folder_path and df must be specified if find_subdir is False'
            self.img_folder_path = img_folder_path
            self.df = df
            self.splits = splits
        else:
            self.pkls = sorted([fname for fname in glob.glob(f'{dir_path}/**/*.pkl', recursive=True)])
            self.csvs = sorted([fname for fname in glob.glob(f'{dir_path}/**/*.csv', recursive=True)])

            if not os.path.exists(self.dir_path):
                raise ValueError('Data folder does not exist.')
            if not self.csvs and not self.pkls:
                raise ValueError(f'No dataset meta information found in {dir_path}')

            if df is not None:
                self.df = df
            else:
                if self.csvs:
                    fp = self.csvs[0]
                    self.df = pd.read_csv(fp)
                elif self.pkls:
                    fp = self.pkls[0]
                    self.df: pd.DataFrame = pd.read_pickle(fp)
                if not self.csvs and not self.pkls:
                    raise ValueError(f'No dataset meta information found in {dir_path}')

            if splits is not None:
                self.splits = splits
            else:
                self.splits = self.find_dataset_splits()

            self.img_folder_path = None
            if img_folder_path is not None:
                self.img_folder_path = img_folder_path
            else:
                for k, v in self.splits.items():
                    if 'path' in v:
                        if v['path']:
                            self.img_folder_path = v['path']
                            break
                if not self.img_folder_path:
                    raise ValueError('No image folder path found')

    def find_dataset_splits(self):
        assert self.find_subdir, 'find_subdir must be True to find dataset splits'
        splits = {
            'train': dict(),
            'val': dict(),
            'test': dict()
        }
        no_datafolder = True
        for split in splits.keys():
            dirs = [fname for fname in glob.glob(f'{self.dir_path}/**/{split}', recursive=True)]
            if dirs:
                no_datafolder = False
                assert len(dirs) == 1, f'found more than 1 folder for {split} split'
                if os.path.isdir(dirs[0]):
                    splits[split]['path'] = dirs[0]
            if self.csvs:
                df_fp = [fp for fp in self.csvs if split in fp.split('/')[-1]]
                if df_fp:
                    splits[split]['df'] = pd.read_csv(df_fp[0])
            elif self.pkls:
                df_fp = [fp for fp in self.pkls if split in fp.split('/')[-1]]
                if df_fp:
                    splits[split]['df'] = pd.read_pickle(df_fp[0])
        if no_datafolder:
            raise ValueError(f'No dataset splits found in data folder {self.dir_path}')
        return splits

    def get_object(self, obj, obj_size: Tuple[int, int], mode: str) -> NDArray:
        raise NotImplementedError


def get_grid(grid_size):
    # return (grid_size,grid_size) array of sg.space.values
    # convert from grid to space

    x_space, y_space = grid_size[0] + 1, grid_size[1] + 1
    x_coords, y_coords = np.linspace(0, 1, x_space), np.linspace(0, 1, y_space)
    xx, yy = np.meshgrid(x_coords, y_coords, sparse=True)
    xx, yy = xx.flatten(), yy.flatten()
    grid_spaces = {(i, j): [(x_i, x_k), (y_i, y_k)] for i, (x_i, x_k) in enumerate(zip(xx[0::], xx[1::])) for
                   j, (y_i, y_k) in enumerate(zip(yy[0::], yy[1::]))}
    return OrderedDict(grid_spaces)



"""
abstract base classes for any environment,
see class methods for their expected behaviours
"""

from collections import defaultdict
from typing import Tuple, Dict, List, Iterable, Any, Union

import networkx as nx
import numpy as np

from iwisdm.envs.registration import EnvSpec, StimData, Constant


class Attribute(object):
    """Base class for attributes."""

    env_spec: EnvSpec = None
    _stim_data: StimData = None
    constants: Constant = None

    def __init__(self, value):
        if isinstance(value, np.int_):
            value = int(value)
        elif isinstance(value, list):
            value = tuple(value)
        self.value = value
        self.parent = list()

    def __call__(self, *args):
        """Including a call function to be consistent with Operator class."""
        return self

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        """Override the default Equals behavior."""
        if isinstance(other, self.__class__):
            return self.value == other.value
        return False

    def __ne__(self, other):
        """Define a non-equality test."""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return True

    def __hash__(self):
        return hash(self.value)

    def self_json(self):
        return {}

    def to_json(self):
        info = dict()
        info['name'] = self.__class__.__name__
        info['value'] = self.value
        info.update(self.self_json())
        return info

    def sample(self):
        raise NotImplementedError('Abstract method')

    def resample(self, attr):
        raise NotImplementedError('Abstract method')

    def has_value(self):
        return self.value is not None


class Stimulus(object):
    """A stimulus on the screen.

    A stimulus is a collection of attributes.
    """
    env_spec: EnvSpec = None
    _stim_data: StimData = None
    constants: Constant = None
    epoch: List[int] = None

    def __str__(self):
        raise NotImplementedError

    def change_attr(self, attr: Attribute):
        raise NotImplementedError()

    def compare_attrs(self, other, attrs: List[str] = None):
        raise NotImplementedError

    def dump(self):
        """Returns representation of self suitable for dumping as json."""
        raise NotImplementedError

    def to_static(self):
        """Convert self to a list of StaticObjects."""
        raise NotImplementedError

    def merge(self, obj):
        raise NotImplementedError

    def copy(self):
        raise NotImplementedError

    def render(self, *args):
        """
        return the rendered image of the stimulus
        """
        raise NotImplementedError


class StimuliSet(object):
    """A collection of objects."""
    env_spec: EnvSpec = None
    _stim_data: StimData = None
    constants: Constant = None

    def __init__(self, n_epoch):
        """Initialize the collection of objects.

        Args:
          n_epoch: int, the number of epochs or frames in the object set
        """
        self.n_epoch = n_epoch
        self.set = list()
        self.end_epoch = list()
        self.dict = defaultdict(list)  # key: epoch, value: list of obj

        self.last_added_obj = None  # Last added object
        self.location = None

    def __iter__(self):
        return self.set.__iter__()

    def __str__(self):
        return '\n'.join([str(o) for o in self])

    def __len__(self):
        return len(self.set)

    def copy(self):
        raise NotImplementedError

    def increase_epoch(self, new_n_epoch):
        """
        increase the number of epochs of the objset by initializing the new epoch indices with empty lists
        :param new_n_epoch: new number of epochs
        :return:
        """
        for i in range(self.n_epoch, new_n_epoch):
            self.dict[i] = list()
        self.n_epoch = new_n_epoch
        return

    def add(self,
            stim: Stimulus,
            epoch_now,
            add_if_exist=False,
            delete_if_can=True,
            merge_idx=None,
            ):
        """Add an object at the current epoch

        This function will attempt to add the stim if possible.
        It will not only add the object to the objset, but also instantiate the
        attributes such as color, shape, and location if not already instantiated.

        Args:
          stim: a Stimulus instance
          epoch_now: the current epoch when this object is added
          add_if_exist: if True, add object anyway. If False, do not add object if
            already exist
          delete_if_can: Boolean. If True, will delete object if it conflicts with
            current object to be added. Should be set to True for most situations.
          merge_idx: the absolute epoch for adding the object, used for merging task_info
            when lastk of different tasks results in ambiguous epoch_idx
        Returns:
          stim: the added object if object added. The existing object if not added.

        Raises:
          ValueError: if can't find place to put stimuli
        """

        raise NotImplementedError

    def delete(self, stim):
        """Delete an object."""
        i = self.set.index(stim)
        self.set.pop(i)
        self.end_epoch.pop(i)

        for epoch in range(stim.epoch[0], stim.epoch[1]):
            self.dict[epoch].remove(stim)

    def select(self, *args):
        raise NotImplementedError

    def select_now(self, epoch_now: int, *args) -> Iterable[Stimulus]:
        """
        return a subset of the stimuli that are present at the current epoch,
        filter the stimuli based on the provided attributes
        """
        raise NotImplementedError


class Operator(object):
    """Base class for task constructors."""
    env_spec: EnvSpec = None
    _stim_data: StimData = None
    constants: Constant = None

    def __init__(self):
        # Whether this operator is the final operator
        self.child = list()
        self.parent = list()

    def __str__(self):
        pass

    def __hash__(self):
        if self.child:
            c_s = sorted(self.child, key=lambda o: o.__class__.__name__)
            return hash(tuple(
                [self.__class__.__name__] +
                [c for c in c_s if isinstance(c, Operator)]
            ))
        else:
            return hash(self.__class__.__name__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if len(self.child) != len(other.child):
                return False
            else:
                for c, o_c in zip(
                        sorted(self.child, key=lambda o: o.__class__.__name__),
                        sorted(other.child, key=lambda o: o.__class__.__name__)
                ):
                    if isinstance(c, Operator) and isinstance(o_c, Operator):
                        if c != o_c:
                            return False
                return True
        return False

    def __call__(self, objset, epoch_now):
        del objset
        del epoch_now

    def copy(self):
        raise NotImplementedError

    def set_child(self, child):
        """Set operators as children."""
        try:
            child.parent.append(self)
            self.child.append(child)
        except AttributeError:
            for c in child:
                self.set_child(c)

    def get_expected_input(self, *args, **kwargs):
        """Guess and update the objset at this epoch.

        Args:
          should_be: the supposed output
        """
        raise NotImplementedError('get_expected_input method is not defined.')

    def child_json(self):
        return [c.to_json() for c in self.child]

    def self_json(self):
        return {}

    def to_json(self):
        # return the dictionary for storing into json
        info = dict()
        info['name'] = self.__class__.__name__
        info['child'] = self.child_json()
        info.update(self.self_json())
        return info

    def check_attrs(self):
        """
        check if any children are operators
        :return: True if children contain no operators
        """
        for c in self.child:
            if isinstance(c, Operator):
                return False
        return True

    def get_children_targets(self, objset):
        """
        get the answers to each child recursively
        @param objset:
        @return:
        """
        if not isinstance(self, Operator):
            return
        target = str(self(objset, objset.n_epoch - 1))
        self_dict = dict()
        self_dict['question'] = str(self)
        self_dict['answer'] = target
        if not self.child:
            return self_dict
        self_dict['children'] = [
            c.get_children_targets(objset)
            for c in self.child if isinstance(c, Operator) and not isinstance(c, Attribute)
        ]
        return self_dict


class Task(object):
    """Base class for tasks."""
    env_spec: EnvSpec = None
    _stim_data: StimData = None
    constants: Constant = None

    def __init__(self, operator=None):
        if operator is None:
            self._operator = Operator()
        else:
            if not isinstance(operator, Operator):
                raise TypeError('operator is the wrong type ' + str(type(operator)))
            self._operator = operator

    def __call__(self, objset, epoch_now):
        # when you want to get the answer to the ask
        return self._operator(objset, epoch_now)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._operator == other._operator
        return False

    def __hash__(self):
        return hash(self._operator)

    def __str__(self):
        return str(self._operator)

    def _get_all_nodes(self, op, visited):
        # used for topological sort, not need to read
        """Get the total number of operators in the graph starting with op."""
        visited[op] = True
        all_nodes = [op]
        for c in op.child:
            if isinstance(c, Operator) and not visited[c]:
                all_nodes.extend(self._get_all_nodes(c, visited))
        return all_nodes

    @property
    def _all_nodes(self):
        """Return all nodes in a list."""
        visited = defaultdict(lambda: False)
        return self._get_all_nodes(self._operator, visited)

    @property
    def operator_size(self):
        """Return the number of unique operators."""
        return len(self._all_nodes)

    def topological_sort_visit(self, node, visited, stack):
        """Recursive function that visits a root."""

        # Mark the current root as visited.
        visited[node] = True

        # Recur for all the vertices adjacent to this vertex
        for child in node.child:
            if isinstance(child, Operator) and not visited[child]:
                self.topological_sort_visit(child, visited, stack)

        # Push current vertex to stack which stores result
        stack.insert(0, node)

    def topological_sort(self) -> List[Operator]:
        """Perform a topological sort."""
        nodes = self._all_nodes

        # Mark all the vertices as not visited
        visited = defaultdict(lambda: False)
        stack = []

        # Call the recursive helper function to store Topological
        # Sort starting from all vertices one by one
        for node in nodes:
            if not visited[node]:
                self.topological_sort_visit(node, visited, stack)

        # Print contents of stack
        return stack

    def guess_objset(self, *kwargs):
        """
        main function for generating frames based on task graph structure
        iterate through each node in topological order, and propagate the expected inputs from
        predecessors to the successors
        :return: the updated objset after going through the task graph
        """
        raise NotImplementedError

    @property
    def instance_size(self):
        """Return the total number of possible instantiated tasks."""
        raise NotImplementedError('instance_size is not defined for this task.')

    def get_target(self, objset):
        return [self(objset, objset.n_epoch - 1)]

    def get_subtask_targets(self, objset):
        """
        get the answers to each subtask
        @param objset:
        @return: nested lists of answers to each children operator
        """
        targets = [str(target) for target in self.get_target(objset)]
        self_dict = dict()
        self_dict['question'] = str(self)
        self_dict['answer'] = targets[0]
        if not self._operator.child:
            return self_dict
        self_dict['children'] = [
            c.get_children_targets(objset)
            for c in self._operator.child if isinstance(c, Operator)
        ]
        return self_dict

    def is_bool_output(self) -> bool:
        raise NotImplementedError


GRAPH_TUPLE = Tuple[nx.DiGraph, int, int]
TASK = Tuple[Union[Operator, Attribute], Task]


class Env(object):
    """
    Base class for any environment.


    """
    metadata: Dict[str, Any] = dict()
    env_spec: EnvSpec = None
    task_gen_config: Dict[str, Any] = dict()
    constants: Constant = None
    stim_data: StimData = None

    @staticmethod
    def init_stim_data(dataset_fp: str, *args) -> StimData:
        """
        initialize the StimData for a specific environment
        @param dataset_fp:
        @param args:
        @return:
        """
        raise NotImplementedError

    @staticmethod
    def init_env_spec(*args):
        raise NotImplementedError

    def set_stim_data(self, stim_data: StimData):
        raise NotImplementedError

    def set_env_spec(self, env_spec: EnvSpec):
        raise NotImplementedError

    def generate_tasks(self, n: int = 1, *args) -> List[Tuple[GRAPH_TUPLE, TASK]]:
        """
        generate random tasks based on the environment specification
        @param n: how many tasks to generate
        @param args:
        @return: List of tasks, each task is a tuple of
        [
        1. task graph tuple [nx.DiGraph, operator count, root operator index],
        2. task tuple [Operator/Attribute, Task]
        ]
        """
        raise NotImplementedError

    def generate_trials(self,
                        tasks: List[Task],
                        mode: str,
                        **kwargs) -> List[Tuple[List[np.ndarray], List[Dict], Dict]]:
        """
        generate trials based on the tasks and which dataset split to sample stimuli from

        @param tasks: List of tasks to generate trials from
        @param mode: the dataset split [train/valid/test]
        @return: List of trials, each trial is a tuple of
        [List of images, Info dictionaries per task (in case of compositional tasks), Compositional task info dict]
        """
        raise NotImplementedError

    def render_trials(self):
        raise NotImplementedError



import os
from typing import Tuple, List, Union
import shutil
from pathlib import Path

import json
import cv2
import numpy as np
from numpy.typing import NDArray

from iwisdm.core import StimuliSet, StimData


def read_img(fp: str, obj_size: Tuple[int, int], color_format='RGB'):
    image = cv2.imread(fp)
    # image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB) if color_format == 'RGB' else image
    object_arr = cv2.resize(image, obj_size)
    return object_arr


def add_cross(canvas: NDArray, cue_size: float = 0.05):
    """
    Add a cross to the center of the canvas
    @param canvas: the image array
    @param cue_size: the size of the cross relative to the canvas size
    @return:
    """
    img_size = canvas.shape[0]
    radius = int(cue_size * img_size)
    center = (canvas.shape[0] // 2, canvas.shape[1] // 2)
    thickness = int(0.02 * img_size)
    cv2.line(canvas, (center[0] - radius, center[1]),
             (center[0] + radius, center[1]), (255, 255, 255), thickness)
    cv2.line(canvas, (center[0], center[1] - radius),
             (center[0], center[1] + radius), (255, 255, 255), thickness)
    return canvas


def render_stim(canvas, obj, img_size, stim_data: StimData):
    """Render a static object.

    Args:
        canvas: numpy array of type int8 (img_size, img_size, 3). Modified in place.
          Importantly, opencv default is (B, G, R) instead of (R,G,B)
        obj: StaticObject instance
        img_size: int, image size.
        stim_data:
        mode:
    Returns:
        canvas: the modified frame, numpy array of type int8 (img_size, img_size, 3)
    """
    # Fixed specifications, see Space.sample()
    # when sampling, use a 1x1 grid, the most top-left position is (0.1, 0.1),
    # most bottom-right position is (0.9,0.9)
    # changing scaling requires changing space.sample)
    radius = int(0.25 * img_size)

    # Note that OpenCV color is (Blue, Green, Red)
    # TODO: change this from manual to grid_size
    center = [0, 0]
    if obj.location[0] < 0.5:
        center[0] = 56
    else:
        center[0] = 168
    if obj.location[1] < 0.5:
        center[1] = 56
    else:
        center[1] = 168
    # center = (int(obj.location[0] * img_size), int(obj.location[1] * img_size))

    x_offset, x_end = center[0] - radius, center[0] + radius
    y_offset, y_end = center[1] - radius, center[1] + radius
    obj_size = radius * 2
    shape_net_obj = stim_data.get_object(obj, (obj_size, obj_size))
    assert shape_net_obj.shape[:2] == (y_end - y_offset, x_end - x_offset)

    # note that openCV has x and y axis reversed
    canvas[y_offset:y_end, x_offset:x_end] = shape_net_obj
    return canvas


def render_stimset(stim_set: Union[List[StimuliSet], StimuliSet], canvas_size=224, stim_data: StimData = None):
    """
    Render a movie by epoch.

    @param stim_set: a StimuliSet instance or a list of them
    @param canvas_size: overall size of the rendered image
    @param stim_data: the stimuli dataset
    @return: numpy array (n_time, img_size, img_size, 3)
    """
    if not isinstance(stim_set, list):
        stim_set = [stim_set]

    n_objset = len(stim_set)
    n_epoch_max = max([s.n_epoch for s in stim_set])

    # It's faster if use uint8 here, but later conversion to float32 seems slow
    movie = np.zeros((n_objset * n_epoch_max, canvas_size, canvas_size, 3), np.uint8)

    i_frame = 0
    for s in stim_set:
        for epoch_now in range(n_epoch_max):
            canvas = movie[i_frame:i_frame + 1, ...]  # return a view
            canvas = np.squeeze(canvas, axis=0)

            subset = s.select_now(epoch_now)
            for obj in subset:
                canvas = render_stim(canvas, obj.to_static()[0], canvas_size, stim_data)
            i_frame += 1
    return movie


def write_task(task, save_dir_fp, task_id=None):
    """
    Write the task to a json file

    @param task: a TemporalTask instance
    @param save_dir_fp: the directory to save the task
    @return:
    """
    if task_id:
        save_fp = os.path.join(save_dir_fp, f'task_{task_id}.json')
    else:
        save_fp = save_dir_fp
    info = task.to_json()
    with open(save_fp, 'w') as f:
        json.dump(info, f, indent=4)
    return


def write_trial(imgs, compo_info_dict, trial_fp: str) -> None:
    """
    write the trial images, and save the task information in task_info.json

    @param imgs: a list of numpy arrays, each array is an image
    @param compo_info_dict: a dictionary containing task information
    @param trial_fp: the directory to write the frames, usually folder name is trial_i
    @return:
    """

    frames_fp = os.path.join(trial_fp, 'frames')
    if os.path.exists(frames_fp):
        shutil.rmtree(frames_fp)
    os.makedirs(frames_fp)

    for i, img_arr in enumerate(imgs):
        filename = os.path.join(frames_fp, f'epoch{i}.png')
        cv2.imwrite(filename, img_arr)

    filename = os.path.join(frames_fp, 'task_info.json')
    with open(filename, 'w') as f:
        json.dump(compo_info_dict, f, indent=4)
    return


def find_data_folder(data_folder: str = None):
    """
    In the project director data folder,
    find a subdirectory with stimuli images and csv file containing stimuli information.
    The subdirectory should have format:
    project/
        data/
            dataset_name/
                {train, val, test}/
                    ...
                meta.{csv, pkl}
    @param data_folder: the path to the data folder
    @return: the path to the data folder
    """
    if not data_folder:
        data_folder = os.path.join(os.getcwd(), 'data')
    if os.path.isdir(data_folder):
        for sub_dir in os.listdir(data_folder):
            dir_path = os.path.join(data_folder, sub_dir)
            if os.path.isdir(dir_path):
                if list(Path(dir_path).rglob('*.csv')) or list(Path(dir_path).rglob('*.pkl')):
                    if (os.path.isdir(os.path.join(dir_path, 'train')) or os.path.isdir(
                            os.path.join(dir_path, 'validation')) or os.path.isdir(os.path.join(dir_path, 'test'))):
                        return dir_path
    raise ValueError(f'No dataset found in data folder {data_folder}. Specify the dataset path.')
